"""Python bindings for C library for finding and handling crystal."""
# Copyright (C) 2015 Atsushi Togo
# All rights reserved.
#
# This file is part of spglib.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the spglib project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import dataclasses
import sys
import warnings
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Union

import numpy as np

try:
    from . import _spglib  # type: ignore[attr-defined]
except ImportError:
    if sys.version_info < (3, 10):
        from importlib_resources import as_file, files
        from typing_extensions import TypeAlias
    else:
        from importlib.resources import as_file, files
        from typing import TypeAlias
    from ctypes import cdll

    root = files("spglib.lib")
    for file in root.iterdir():
        if "symspg." in file.name:
            with as_file(file) as bundled_lib:
                try:
                    cdll.LoadLibrary(str(bundled_lib))
                    from . import _spglib  # type: ignore[attr-defined]

                    break
                except ImportError as err:
                    raise FileNotFoundError(
                        "Could not load bundled Spglib C library"
                    ) from err
    else:
        raise FileNotFoundError(
            "Spglib C library is not installed and no bundled version was detected"
        )

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Any

    from numpy.typing import ArrayLike, NDArray

warnings.filterwarnings(
    "module", category=DeprecationWarning, message=r"dict interface.*"
)

Lattice: TypeAlias = Sequence[Sequence[float]]
Positions: TypeAlias = Sequence[Sequence[float]]
Numbers: TypeAlias = Sequence[int]
Magmoms: TypeAlias = Union[Sequence[float], Sequence[Sequence[float]]]
Cell: TypeAlias = Union[
    tuple[Lattice, Positions, Numbers], tuple[Lattice, Positions, Numbers, Magmoms]
]


class SpglibError:
    """Error message why spglib failed."""

    message = "no error"


spglib_error = SpglibError()


@dataclasses.dataclass(eq=True, frozen=True)
class DictInterface(Mapping[str, "Any"]):
    """Base class for dataclass with dict interface.

    .. versionadded:: 2.5.0
    .. deprecated:: 2.5.0
        Dict-like interface (``obj['field']``) are deprecated.
        Please use attribute interface instead (``obj.field``)
    """

    def __getitem__(self, key: str) -> Any:
        """Return the value of the key."""
        warnings.warn(
            f"dict interface ({self.__class__.__name__}['{key}']) is deprecated."
            "Use attribute interface ({self.__class__.__name__}.{key}) instead",
            DeprecationWarning,
        )
        return dataclasses.asdict(self)[key]

    def __len__(self) -> int:
        """Return the number of fields."""
        return len(dataclasses.fields(self))

    def __iter__(self) -> Iterator[str]:
        """Return an iterator over the keys."""
        return iter(dataclasses.asdict(self))


@dataclasses.dataclass(eq=False, frozen=True)
class SpglibDataset(DictInterface):
    """Spglib dataset information.

    .. versionadded:: 1.9.4
        The member 'choice' is added.
    .. versionadded:: 2.5.0
    """

    number: int
    """International space group number"""
    hall_number: int
    """Hall number.

    This number is used in
        :func:`get_symmetry_from_database` and
        :func:`get_spacegroup_type`.
    """
    international: str
    """International symbol"""
    hall: str
    """Hall symbol"""
    choice: str
    """Centring, origin, basis vector setting"""
    transformation_matrix: NDArray[np.double]
    """Transformation matrix from input lattice to standardized lattice.

    shape=(3, 3), order='C', dtype='double'

    .. code-block::

        L^original = L^standardized * Tmat.

    See the detail at :ref:`dataset_origin_shift_and_transformation`.
    """
    origin_shift: NDArray[np.double]
    """Origin shift from standardized to input origin.

    shape=(3,), dtype='double'

    See the detail at :ref:`dataset_origin_shift_and_transformation`.
    """
    rotations: NDArray[np.intc]
    """Matrix (rotation) parts of space group operations.

    shape=(n_operations, 3, 3), order='C', dtype='intc'

    .. note::

        Space group operations are obtained by

        .. code-block:: python

            [(r,t) for r, t in zip(rotations, translations)]

    See also :func:`get_symmetry`.
    """
    translations: NDArray[np.double]
    """Vector (translation) parts of space group operations.

    shape=(n_operations, 3), order='C', dtype='double'

    .. note::

        Space group operations are obtained by

        .. code-block:: python

            [(r,t) for r, t in zip(rotations, translations)]

    See also :func:`get_symmetry`.
    """
    wyckoffs: list[str]
    """Wyckoff letters corresponding to the space group type."""
    site_symmetry_symbols: list[str]
    """Site symmetry symbols corresponding to the space group type."""
    crystallographic_orbits: NDArray[np.intc]
    """Symmetrically equivalent atoms, where 'symmetrically' means the space group
    operations corresponding to the space group type.

    This is very similar to ``equivalent_atoms``. See the difference explained in
    ``equivalent_atoms``

    shape=(n_atoms,), dtype='intc'
    """
    equivalent_atoms: NDArray[np.intc]
    """Symmetrically equivalent atoms, where 'symmetrically' means found symmetry
    operations.

    shape=(n_atoms,), dtype='intc'

    In spglib, symmetry operations are given for the input cell.
    When a non-primitive cell is inputted and the lattice made by the non-primitive
    basis vectors breaks its point group, the found symmetry operations may not
    correspond to the crystallographic space group type.
    """
    # Primitive cell
    primitive_lattice: NDArray[np.double]
    """Basis vectors a, b, c of a primitive cell in row vectors.

    shape=(3, 3), order='C', dtype='double'

    This is the primitive cell found inside spglib before applying standardization.
    Therefore, the values are distorted with respect to found space group type.
    """
    mapping_to_primitive: NDArray[np.intc]
    """Atom index mapping from original cell to the primitive cell of
    ``primitive_lattice``.

    shape=(n_atoms), dtype='intc'
    """
    # Idealized standardized unit cell
    std_lattice: NDArray[np.double]
    """Basis vectors a, b, c of a standardized cell in row vectors.

    shape=(3, 3), order='C', dtype='double'
    """
    std_positions: NDArray[np.double]
    """Positions of atoms in the standardized cell in fractional coordinates.

    shape=(n_atoms, 3), order='C', dtype='double'
    """
    std_types: NDArray[np.intc]
    """Identity numbers of atoms in the standardized cell.

    shape=(n_atoms,), dtype='intc'
    """
    std_rotation_matrix: NDArray[np.double]
    """Rigid rotation matrix to rotate from standardized basis vectors to idealized
    standardized orthonormal basis vectors.

    shape=(3, 3), order='C', dtype='double'

    .. code-block::

        L^idealized = R * L^standardized.

    See the detail at :ref:`dataset_idealized_cell`.
    """
    std_mapping_to_primitive: NDArray[np.intc]
    """``std_positions`` index mapping to those of atoms of the primitive cell in the
    standardized cell."""
    pointgroup: str
    """Pointgroup symbol in Hermann-Mauguin notation."""


@dataclasses.dataclass(eq=False, frozen=True)
class SpglibMagneticDataset(DictInterface):
    """Spglib magnetic dataset information.

    See :ref:`magnetic_spglib_dataset` in detail.

    .. versionadded:: 2.5.0
    """

    # Magnetic space-group type
    uni_number: int
    """UNI number between 1 to 1651"""
    msg_type: int
    """Magnetic space groups (MSG) is classified by its family space group (FSG) and
    maximal space subgroup (XSG).

    FSG is a non-magnetic space group obtained by ignoring time-reversal term in MSG.
    XSG is a space group obtained by picking out non time-reversal operations in MSG.

    - msg_type==1 (type-I):
        MSG, XSG, FSG are all isomorphic.
    - msg_type==2 (type-II):
        XSG and FSG are isomorphic, and MSG is generated from XSG and pure time reversal
        operations.
    - msg_type==3 (type-III):
        XSG is a proper subgroup of MSG with isomorphic translational subgroups.
    - msg_type==4 (type-IV):
        XSG is a proper subgroup of MSG with isomorphic point group.
    """
    hall_number: int
    """For type-I, II, III, Hall number of FSG; for type-IV, that of XSG"""
    tensor_rank: int
    """Rank of magmoms."""

    # Magnetic symmetry operations
    n_operations: int
    """Number of magnetic symmetry operations"""
    rotations: NDArray[np.intc]
    """Rotation (matrix) parts of symmetry operations

    shape: (n_operations, 3, 3)
    """
    translations: NDArray[np.double]
    """Translation (vector) parts of symmetry operations

    shape: (n_operations, 3)
    """
    time_reversals: NDArray[np.intc]
    """Time reversal part of magnetic symmetry operations.

    True indicates time reversal operation, and False indicates an ordinary operation.

    shape: (n_operations, )
    """

    # Equivalent atoms
    n_atoms: int
    """Number of atoms in the input cell"""
    equivalent_atoms: NDArray[np.intc]
    """Symmetrically equivalent atoms, where 'symmetrically' means found symmetry
    operations.
    """

    # Transformation to standardized setting
    transformation_matrix: NDArray[np.intc]
    """Transformation matrix from input lattice to standardized

    shape: (3, 3)
    """
    origin_shift: NDArray[np.double]
    """Origin shift from standardized to input origin

    shape: (3, )
    """

    # Standardized crystal structure
    n_std_atoms: int
    """Number of atoms in standardized unit cell"""
    std_lattice: NDArray[np.double]
    """Row-wise lattice vectors of the standardized unit cell

    shape: (3, 3)
    """
    std_types: NDArray[np.intc]
    """Identity numbers of atoms in the standardized unit cell

    shape: (n_std_atoms, )
    """
    std_positions: NDArray[np.double]
    """Fractional coordinates of atoms in the standardized unit cell

    shape: (n_std_atoms, 3)
    """
    std_tensors: NDArray[np.double]
    """
    shape:
        (n_std_atoms, ) for collinear magnetic moments.
        (n_std_atoms, 3) for vector non-collinear magnetic moments.
    """
    std_rotation_matrix: NDArray[np.double]
    """Rigid rotation matrix to rotate from standardized basis vectors to idealized
    standardized basis vectors"""

    # Intermediate data in symmetry search
    primitive_lattice: NDArray[np.double]
    """Basis vectors of primitive lattice.

    shape: (3, 3)
    """


@dataclasses.dataclass(eq=True, frozen=True)
class SpaceGroupType(DictInterface):
    """Space group type information.

    More details are found at :ref:`spglib-dataset`.

    .. versionchanged:: 2.0
        ``hall_number`` member is added.
    .. versionadded:: 2.5.0
    """

    number: int
    """International space group number"""
    international_short: str
    """International short symbol"""
    international_full: str
    """International full symbol"""
    international: str
    """International symbol"""
    schoenflies: str
    """Schoenflies symbol"""
    hall_number: int
    """Hall symbol serial number"""
    hall_symbol: str
    """Hall symbol"""
    choice: str
    """Centring, origin, basis vector setting"""
    pointgroup_international: str
    """International symbol of crystallographic point group"""
    pointgroup_schoenflies: str
    """Schoenflies symbol of crystallographic point group"""
    arithmetic_crystal_class_number: int
    """Arithmetic crystal class number"""
    arithmetic_crystal_class_symbol: str
    """Arithmetic crystal class symbol"""


@dataclasses.dataclass(eq=True, frozen=True)
class MagneticSpaceGroupType(DictInterface):
    """Magnetic space group type information.

    .. versionadded:: 2.5.0
    """

    uni_number: int
    """Serial number of UNI (or BNS) symbols"""
    litvin_number: int
    """Serial number in Litvin's [Magnetic group tables](https://www.iucr.org/publ/978-0-9553602-2-0)"""
    bns_number: str
    """BNS number e.g. '151.32'"""
    og_number: str
    """OG number e.g. '153.4.1270'"""
    number: int
    """ITA's serial number of space group for reference setting"""
    type: int
    """Type of MSG from 1 to 4"""


def get_version():
    """Return version number of spglib with tuple of three numbers.

    .. versionadded:: 1.8.3
    .. deprecated:: 2.3.0
        Use :py:func:`spg_get_version` and ``spglib.__version__`` instead
    """
    warnings.warn(
        "get_version() is deprecated. Use __version__ for the python binding"
        "version and get_spg_version for the detected spglib library version.",
        DeprecationWarning,
    )
    _set_no_error()
    return _spglib.version_tuple()


def spg_get_version():
    """Get the X.Y.Z version of the detected spglib C library.

    .. versionadded:: 2.3.0
    :return: version string
    """
    _set_no_error()
    return _spglib.version_string()


def spg_get_version_full():
    """Get the full version of the detected spglib C library.

    .. versionadded:: 2.3.0
    :return: full version string
    """
    _set_no_error()
    return _spglib.version_full()


def spg_get_commit():
    """Get the commit of the detected spglib C library.

    .. versionadded:: 2.3.0
    :return: commit string
    """
    _set_no_error()
    return _spglib.commit()


def get_symmetry(
    cell: Cell,
    symprec=1e-5,
    angle_tolerance=-1.0,
    mag_symprec=-1.0,
    is_magnetic=True,
) -> dict | None:
    r"""Find symmetry operations from a crystal structure and site tensors.

    .. warning::
        :func:`get_symmetry` with ``is_magnetic=True`` is deprecated at version 2.0.

    Use :func:`get_magnetic_symmetry` for magnetic symmetry search.

    Parameters
    ----------
    cell : tuple
        Crystal structure given in tuple.
        It has to follow the following form,
        (basis vectors, atomic points, types in integer numbers, ...)

        - basis vectors : array_like
            shape=(3, 3), order='C', dtype='double'

            .. code-block:: python

                [[a_x, a_y, a_z],
                [b_x, b_y, b_z],
                [c_x, c_y, c_z]]

        - atomic points : array_like
            shape=(num_atom, 3), order='C', dtype='double'

            Atomic position vectors with respect to basis vectors, i.e.,
            given in  fractional coordinates.
        - types : array_like
            shape=(num_atom, ), dtype='intc'

            Integer numbers to distinguish species.
        - optional data :
            case-I: Scalar
                shape=(num_atom, ), dtype='double'

                Each atomic site has a scalar value. With is_magnetic=True,
                values are included in the symmetry search in a way of
                collinear magnetic moments.
            case-II: Vectors
                shape=(num_atom, 3), order='C', dtype='double'

                Each atomic site has a vector. With is_magnetic=True,
                vectors are included in the symmetry search in a way of
                non-collinear magnetic moments.
    symprec : float
        Symmetry search tolerance in the unit of length.
    angle_tolerance : float
        Symmetry search tolerance in the unit of angle deg.
        Normally it is not recommended to use this argument.
        See a bit more detail at :ref:`variables_angle_tolerance`.
        If the value is negative, an internally optimized routine is used to
        judge symmetry.
    mag_symprec : float
        Tolerance for magnetic symmetry search in the unit of magnetic moments.
        If not specified, use the same value as symprec.
    is_magnetic : bool
        When optional data (4th element of cell tuple) is given in case-II,
        the symmetry search is performed considering magnetic symmetry, which
        may be corresponding to that for non-collinear calculation. Default is
        True, but this does nothing unless optional data is supplied.

    Returns
    -------
    symmetry: dict
        Rotation parts and translation parts of symmetry operations are represented
        with respect to basis vectors.
        When the search failed, :code:`None` is returned.

        - 'rotations' : ndarray
            shape=(num_operations, 3, 3), order='C', dtype='intc'

            Rotation (matrix) parts of symmetry operations
        - 'translations' : ndarray
            shape=(num_operations, 3), dtype='double'

            Translation (vector) parts of symmetry operations
        - 'time_reversals': ndarray (exists when the optional data is given)
            shape=(num_operations, ), dtype='bool\_'

            Time reversal part of magnetic symmetry operations.
            True indicates time reversal operation, and False indicates
            an ordinary operation.
        - 'equivalent_atoms' : ndarray
            shape=(num_atoms, ), dtype='intc'

            A mapping table of atoms to symmetrically independent atoms.
            This is used to find symmetrically equivalent atoms.
            The numbers contained are the indices of atoms starting from 0,
            i.e., the first atom is numbered as 0, and
            then 1, 2, 3, ... :code:`np.unique(equivalent_atoms)` gives representative
            symmetrically independent atoms. A list of atoms that are
            symmetrically equivalent to some independent atom (here for example 1
            is in :code:`equivalent_atom`) is found by
            :code:`np.where(equivalent_atom=1)[0]`.

    Notes
    -----
    The orders of the rotation matrices and the translation
    vectors correspond with each other, e.g. , the second symmetry
    operation is organized by the set of the second rotation matrix and second
    translation vector in the respective arrays. Therefore a set of
    symmetry operations may obtained by

    .. code-block:: python

        [(r, t) for r, t in zip(dataset['rotations'], dataset['translations'])]

    The operations are given with respect to the fractional coordinates
    (not for Cartesian coordinates). The rotation matrix and translation
    vector are used as follows:

    .. code-block::

        new_vector[3x1] = rotation[3x3] * vector[3x1] + translation[3x1]

    The three values in the vector are given for the a, b, and c axes,
    respectively.

    """
    _set_no_error()

    _, _, _, magmoms = _expand_cell(cell)

    if magmoms is None:
        # Get symmetry operations without on-site tensors (i.e. normal crystal)
        dataset = get_symmetry_dataset(
            cell,
            symprec=symprec,
            angle_tolerance=angle_tolerance,
        )
        if dataset is None:
            _set_error_message()
            return None

        return {
            "rotations": dataset.rotations,
            "translations": dataset.translations,
            "equivalent_atoms": dataset.equivalent_atoms,
        }
    else:
        warnings.warn(
            "Use get_magnetic_symmetry() for cell with magnetic moments.",
            DeprecationWarning,
        )
        return get_magnetic_symmetry(
            cell,
            symprec=symprec,
            angle_tolerance=angle_tolerance,
            mag_symprec=mag_symprec,
            is_axial=None,
            with_time_reversal=is_magnetic,
        )


def get_magnetic_symmetry(
    cell: Cell,
    symprec=1e-5,
    angle_tolerance=-1.0,
    mag_symprec=-1.0,
    is_axial=None,
    with_time_reversal=True,
) -> dict | None:
    r"""Find magnetic symmetry operations from a crystal structure and site tensors.

    Parameters
    ----------
    cell : tuple
        Crystal structure given either in tuple.
        In the case given by a tuple, it has to follow the form below,

        (basis vectors, atomic points, types in integer numbers, ...)

        - basis vectors : array_like
            shape=(3, 3), order='C', dtype='double'

            .. code-block::

                [[a_x, a_y, a_z],
                [b_x, b_y, b_z],
                [c_x, c_y, c_z]]

        - atomic points : array_like
            shape=(num_atom, 3), order='C', dtype='double'

            Atomic position vectors with respect to basis vectors, i.e.,
            given in  fractional coordinates.
        - types : array_like
            shape=(num_atom, ), dtype='intc'

            Integer numbers to distinguish species.
        - magmoms:
            case-I: Scalar
                shape=(num_atom, ), dtype='double'

                Each atomic site has a scalar value. With is_magnetic=True,
                values are included in the symmetry search in a way of
                collinear magnetic moments.
            case-II: Vectors
                shape=(num_atom, 3), order='C', dtype='double'

                Each atomic site has a vector. With is_magnetic=True,
                vectors are included in the symmetry search in a way of
                non-collinear magnetic moments.
    symprec : float
        Symmetry search tolerance in the unit of length.
    angle_tolerance : float
        Symmetry search tolerance in the unit of angle deg.
        Normally it is not recommended to use this argument.
        See a bit more detail at :ref:`variables_angle_tolerance`.
        If the value is negative, an internally optimized routine is used to judge
        symmetry.
    mag_symprec : float
        Tolerance for magnetic symmetry search in the unit of magnetic moments.
        If not specified, use the same value as symprec.
    is_axial: None or bool
        Set `is_axial=True` if `magmoms` does not change their sign by improper
        rotations. If not specified, set `is_axial=False` when
        `magmoms.shape==(num_atoms, )`, and set `is_axial=True` when
        `magmoms.shape==(num_atoms, 3)`. These default settings correspond to
        collinear and non-collinear spins.
    with_time_reversal: bool
        Set `with_time_reversal=True` if `magmoms` change their sign by time-reversal
        operations. Default is True.

    Returns
    -------
    symmetry: dict or None
        Rotation parts and translation parts of symmetry operations represented
        with respect to basis vectors and atom index mapping by symmetry
        operations.

        - 'rotations' : ndarray
            shape=(num_operations, 3, 3), order='C', dtype='intc'

            Rotation (matrix) parts of symmetry operations
        - 'translations' : ndarray
            shape=(num_operations, 3), dtype='double'

            Translation (vector) parts of symmetry operations
        - 'time_reversals': ndarray
            shape=(num_operations, ), dtype='bool\_'

            Time reversal part of magnetic symmetry operations.
            True indicates time reversal operation, and False indicates
            an ordinary operation.
        - 'equivalent_atoms' : ndarray
            shape=(num_atoms, ), dtype='intc'
        - 'primitive_lattice': ndarray
            shape=(3, 3), dtype='double'

    Notes
    -----
    .. versionadded:: 2.0

    """
    _set_no_error()

    lattice, positions, numbers, magmoms = _expand_cell(cell)
    if magmoms is None:
        raise TypeError("Specify magnetic moments in cell.")

    max_size = len(positions) * 96
    rotations = np.zeros((max_size, 3, 3), dtype="intc", order="C")
    translations = np.zeros((max_size, 3), dtype="double", order="C")
    equivalent_atoms = np.zeros(len(magmoms), dtype="intc")
    primitive_lattice = np.zeros((3, 3), dtype="double", order="C")
    # (magmoms.ndim - 1) has to be equal to the rank of physical
    # tensors, e.g., ndim=1 for collinear, ndim=2 for non-collinear.
    if magmoms.ndim == 1 or magmoms.ndim == 2:
        spin_flips = np.zeros(max_size, dtype="intc")
    else:
        spin_flips = None

    # Infer is_axial value from tensor_rank to keep backward compatibility
    if is_axial is None:
        if magmoms.ndim == 1:
            is_axial = False  # Collinear spin
        elif magmoms.ndim == 2:
            is_axial = True  # Non-collinear spin

    num_sym = _spglib.symmetry_with_site_tensors(
        rotations,
        translations,
        equivalent_atoms,
        primitive_lattice,
        spin_flips,
        lattice,
        positions,
        numbers,
        magmoms,
        with_time_reversal * 1,
        is_axial * 1,
        symprec,
        angle_tolerance,
        mag_symprec,
    )

    if num_sym == 0:
        _set_error_message()
        return None
    else:
        spin_flips = np.array(spin_flips[:num_sym], dtype="intc", order="C")
        # True for time reversal operation, False for ordinary operation
        time_reversals = spin_flips == -1
        return {
            "rotations": np.array(rotations[:num_sym], dtype="intc", order="C"),
            "translations": np.array(translations[:num_sym], dtype="double", order="C"),
            "time_reversals": time_reversals,
            "equivalent_atoms": equivalent_atoms,
            "primitive_lattice": np.array(
                np.transpose(primitive_lattice),
                dtype="double",
                order="C",
            ),
        }


def _build_dataset_dict(spg_ds: list) -> SpglibDataset:
    letters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"

    dataset = SpglibDataset(
        number=spg_ds[0],
        hall_number=spg_ds[1],
        international=spg_ds[2].strip(),
        hall=spg_ds[3].strip(),
        choice=spg_ds[4].strip(),
        transformation_matrix=np.array(
            spg_ds[5],
            dtype="double",
            order="C",
        ),
        origin_shift=np.array(spg_ds[6], dtype="double"),
        rotations=np.array(spg_ds[7], dtype="intc", order="C"),
        translations=np.array(spg_ds[8], dtype="double", order="C"),
        wyckoffs=[letters[x] for x in spg_ds[9]],
        site_symmetry_symbols=[s.strip() for s in spg_ds[10]],
        crystallographic_orbits=np.array(
            spg_ds[11],
            dtype="intc",
        ),
        equivalent_atoms=np.array(spg_ds[12], dtype="intc"),
        primitive_lattice=np.array(
            np.transpose(spg_ds[13]),
            dtype="double",
            order="C",
        ),
        mapping_to_primitive=np.array(
            spg_ds[14],
            dtype="intc",
        ),
        std_lattice=np.array(
            np.transpose(spg_ds[15]),
            dtype="double",
            order="C",
        ),
        std_types=np.array(spg_ds[16], dtype="intc"),
        std_positions=np.array(spg_ds[17], dtype="double", order="C"),
        std_rotation_matrix=np.array(
            spg_ds[18],
            dtype="double",
            order="C",
        ),
        std_mapping_to_primitive=np.array(
            spg_ds[19],
            dtype="intc",
        ),
        pointgroup=spg_ds[20].strip(),
    )
    return dataset


def get_symmetry_dataset(
    cell: Cell,
    symprec=1e-5,
    angle_tolerance=-1.0,
    hall_number=0,
) -> SpglibDataset | None:
    """Search symmetry dataset from an input cell.

    Parameters
    ----------
    cell, symprec, angle_tolerance:
        See :func:`get_symmetry`.
    hall_number : int
        If a serial number of Hall symbol (>0) is given, the database
        corresponding to the Hall symbol is made.

        The mapping from Hall symbols to a space-group-type is the many-to-one mapping.
        Without specifying this
        option (i.e., in the case of ``hall_number=0``), always the first one
        (the smallest serial number corresponding to the space-group-type in
        [list of space groups (Seto's web site)](https://yseto.net/en/sg/sg1))
        among possible choices and settings is chosen as default. This
        argument is useful when the other choice (or setting) is expected to
        be hooked.

        This affects to the obtained values of `international`,
        `hall`, `choice`, `transformation_matrix`,
        `origin shift`, `wyckoffs`, `std_lattice`, `std_positions`,
        `std_types` and `std_rotation_matrix`, but not to `rotations`
        and `translations` since the later set is defined with respect to
        the basis vectors of user's input (the `cell` argument).

        See also :ref:`dataset_spg_get_dataset_spacegroup_type`.

    Returns
    -------
    dataset: :class:`SpglibDataset` | None
        If it fails, None is returned. Otherwise a dictionary is returned.
        More details are found at :ref:`spglib-dataset`.

    """
    _set_no_error()

    lattice, positions, numbers, _ = _expand_cell(cell)

    spg_ds = _spglib.dataset(
        lattice,
        positions,
        numbers,
        hall_number,
        symprec,
        angle_tolerance,
    )
    if spg_ds is None:
        _set_error_message()
        return None

    dataset = _build_dataset_dict(spg_ds)
    return dataset


def get_symmetry_layerdataset(
    cell: Cell, aperiodic_dir=2, symprec=1e-5
) -> SpglibDataset | None:
    """TODO: Add comments."""
    _set_no_error()

    lattice, positions, numbers, _ = _expand_cell(cell)

    spg_ds = _spglib.layerdataset(
        lattice,
        positions,
        numbers,
        aperiodic_dir,
        symprec,
    )
    if spg_ds is None:
        _set_error_message()
        return None

    dataset = _build_dataset_dict(spg_ds)

    return dataset


def get_magnetic_symmetry_dataset(
    cell: Cell,
    is_axial=None,
    symprec=1e-5,
    angle_tolerance=-1.0,
    mag_symprec=-1.0,
) -> SpglibMagneticDataset | None:
    """Search magnetic symmetry dataset from an input cell. If it fails, return None.

    Parameters
    ----------
    cell, is_axial, symprec, angle_tolerance, mag_symprec:
        See :func:`get_magnetic_symmetry`.

    Returns
    -------
    dataset : :class:`SpglibMagneticDataset` or None

    Notes
    -----
    .. versionadded:: 2.0

    """  # noqa: E501
    _set_no_error()

    lattice, positions, numbers, magmoms = _expand_cell(cell)

    tensor_rank = magmoms.ndim - 1

    # If is_axial is not specified, select collinear or non-collinear spin cases
    if is_axial is None:
        if tensor_rank == 0:
            is_axial = False  # Collinear spin
        elif tensor_rank == 1:
            is_axial = True  # Non-collinear spin

    spg_ds = _spglib.magnetic_dataset(
        lattice,
        positions,
        numbers,
        magmoms,
        tensor_rank,
        is_axial,
        symprec,
        angle_tolerance,
        mag_symprec,
    )
    if spg_ds is None:
        _set_error_message()
        return None

    tensor_rank = spg_ds[3]
    std_tensors = np.array(spg_ds[16], dtype="double", order="C")
    if tensor_rank == 1:
        std_tensors = std_tensors.reshape(-1, 3)

    dataset = SpglibMagneticDataset(
        # Magnetic space-group type
        uni_number=spg_ds[0],
        msg_type=spg_ds[1],
        hall_number=spg_ds[2],
        tensor_rank=tensor_rank,
        # Magnetic symmetry operations
        n_operations=spg_ds[4],
        rotations=np.array(spg_ds[5], dtype="intc", order="C"),
        translations=np.array(spg_ds[6], dtype="double", order="C"),
        time_reversals=np.array(spg_ds[7], dtype="intc"),
        # Equivalent atoms
        n_atoms=spg_ds[8],
        equivalent_atoms=np.array(spg_ds[9], dtype="intc"),
        # Transformation to standardized setting
        transformation_matrix=np.array(spg_ds[10], dtype="double", order="C"),
        origin_shift=np.array(spg_ds[11], dtype="double"),
        # Standardized crystal structure
        n_std_atoms=spg_ds[12],
        std_lattice=np.array(np.transpose(spg_ds[13]), dtype="double", order="C"),
        std_types=np.array(spg_ds[14], dtype="intc"),
        std_positions=np.array(spg_ds[15], dtype="double", order="C"),
        std_tensors=std_tensors,
        std_rotation_matrix=np.array(spg_ds[17], dtype="double", order="C"),
        # Intermediate datum in symmetry search
        primitive_lattice=np.array(np.transpose(spg_ds[18]), dtype="double", order="C"),
    )

    return dataset


def get_layergroup(cell: Cell, aperiodic_dir=2, symprec=1e-5) -> SpglibDataset | None:
    """Return layer group in ....

    If it fails, None is returned.

    """
    _set_no_error()

    dataset = get_symmetry_layerdataset(
        cell,
        aperiodic_dir=aperiodic_dir,
        symprec=symprec,
    )
    return dataset


def get_spacegroup(
    cell: Cell,
    symprec=1e-5,
    angle_tolerance=-1.0,
    symbol_type=0,
) -> str | None:
    """Return space group in international table symbol and number as a string.

    With ``symbol_type=1``, Schoenflies symbol is given instead of international symbol.

    :rtype: str | None
    :return:
        If it fails, None is returned.
    """
    _set_no_error()

    dataset = get_symmetry_dataset(
        cell,
        symprec=symprec,
        angle_tolerance=angle_tolerance,
    )

    if dataset is None:
        _set_error_message()
        return None

    spg_type = get_spacegroup_type(dataset.hall_number)
    if spg_type is None:
        return None

    if symbol_type == 1:
        return "%s (%d)" % (spg_type.schoenflies, dataset.number)
    else:
        return "%s (%d)" % (spg_type.international_short, dataset.number)


def get_spacegroup_type(hall_number: int) -> SpaceGroupType | None:
    """Translate Hall number to space group type information. If it fails, return None.

    This function allows to directly access to the space-group-type database
    in spglib (spg_database.c).
    To specify the space group type with a specific choice, ``hall_number`` is used.
    The definition of ``hall_number`` is found at
    :ref:`dataset_spg_get_dataset_spacegroup_type`.

    :param hall_number: Serial number for Hall symbol
    :return: :class:`SpaceGroupType` or None

    .. versionadded:: 1.9.4
    """
    _set_no_error()

    spg_type_list = _spglib.spacegroup_type(hall_number)

    if spg_type_list is not None:
        spg_type = SpaceGroupType(
            number=spg_type_list[0],
            international_short=spg_type_list[1].strip(),
            international_full=spg_type_list[2].strip(),
            international=spg_type_list[3].strip(),
            schoenflies=spg_type_list[4].strip(),
            hall_number=spg_type_list[5],
            hall_symbol=spg_type_list[6].strip(),
            choice=spg_type_list[7].strip(),
            pointgroup_international=spg_type_list[8].strip(),
            pointgroup_schoenflies=spg_type_list[9].strip(),
            arithmetic_crystal_class_number=spg_type_list[10],
            arithmetic_crystal_class_symbol=spg_type_list[11].strip(),
        )
        return spg_type
    else:
        _set_error_message()
        return None


def get_spacegroup_type_from_symmetry(
    rotations,
    translations,
    lattice=None,
    symprec=1e-5,
) -> SpaceGroupType | None:
    """Return space-group type information from symmetry operations.

    This is expected to work well for the set of symmetry operations whose
    distortion is small. The aim of making this feature is to find space-group-type
    for the set of symmetry operations given by the other source than spglib. The
    parameter ``lattice`` is used as the distance measure for ``symprec``. If this
    is not given, the cubic basis vector whose lengths are one is used.

    Parameters
    ----------
    rotations : array_like
        Matrix parts of space group operations.
        shape=(n_operations, 3, 3), order='C', dtype='intc'
    translations : array_like
        Vector parts of space group operations.
        shape=(n_operations, 3), order='C', dtype='double'
    lattice : array_like, optional
        Basis vectors a, b, c given in row vectors. This is used as the measure of
        distance. Default is None, which gives unit matrix.
        shape=(3, 3), order='C', dtype='double'
    symprec: float
        See :func:`get_symmetry`.

    Returns
    -------
    spacegroup_type : :class:`SpaceGroupType` or None

    Notes
    -----
    .. versionadded:: 2.0

    This is the replacement of :func:`get_hall_number_from_symmetry`.

    """
    r = np.array(rotations, dtype="intc", order="C")
    t = np.array(translations, dtype="double", order="C")
    if lattice is None:
        _lattice = np.eye(3, dtype="double", order="C")
    else:
        _lattice = np.array(lattice, dtype="double", order="C")

    _set_no_error()

    spg_type_list = _spglib.spacegroup_type_from_symmetry(r, t, _lattice, symprec)
    if spg_type_list is not None:
        spg_type = SpaceGroupType(
            number=spg_type_list[0],
            international_short=spg_type_list[1].strip(),
            international_full=spg_type_list[2].strip(),
            international=spg_type_list[3].strip(),
            schoenflies=spg_type_list[4].strip(),
            hall_number=spg_type_list[5],
            hall_symbol=spg_type_list[6].strip(),
            choice=spg_type_list[7].strip(),
            pointgroup_international=spg_type_list[8].strip(),
            pointgroup_schoenflies=spg_type_list[9].strip(),
            arithmetic_crystal_class_number=spg_type_list[10],
            arithmetic_crystal_class_symbol=spg_type_list[11].strip(),
        )
        return spg_type
    else:
        _set_error_message()
        return None


def get_magnetic_spacegroup_type(uni_number: int) -> MagneticSpaceGroupType | None:
    """Translate UNI number to magnetic space group type information.

    If fails, return None.

    Parameters
    ----------
    uni_number : int
        UNI number between 1 to 1651

    Returns
    -------
    magnetic_spacegroup_type: :class:`MagneticSpaceGroupType` | None

    Notes
    -----
    .. versionadded:: 2.0

    """
    _set_no_error()

    msg_type_list = _spglib.magnetic_spacegroup_type(uni_number)

    if msg_type_list is not None:
        msg_type = MagneticSpaceGroupType(
            uni_number=msg_type_list[0],
            litvin_number=msg_type_list[1],
            bns_number=msg_type_list[2].strip(),
            og_number=msg_type_list[3].strip(),
            number=msg_type_list[4],
            type=msg_type_list[5],
        )
        return msg_type
    else:
        _set_error_message()
        return None


def get_magnetic_spacegroup_type_from_symmetry(
    rotations: ArrayLike[np.intc],
    translations: ArrayLike[np.double],
    time_reversals: ArrayLike[np.intc],
    lattice: ArrayLike[np.double] | None = None,
    symprec: float = 1e-5,
) -> MagneticSpaceGroupType | None:
    """Return magnetic space-group type information from symmetry operations.

    Parameters
    ----------
    rotations, translations, time_reversals:
        See returns of :func:`get_magnetic_symmetry`.
    lattice : (Optional) array_like (3, 3)
        Basis vectors a, b, c given in row vectors. This is used as the measure of
        distance. Default is None, which gives unit matrix.
    symprec: float
        See :func:`get_symmetry`.

    Returns
    -------
    magnetic_spacegroup_type: :class:`MagneticSpaceGroupType` | None

    """
    rots = np.array(rotations, dtype="intc", order="C")
    trans = np.array(translations, dtype="double", order="C")
    timerev = np.array(time_reversals, dtype="intc", order="C")
    if lattice is None:
        latt = np.eye(3, dtype="double", order="C")
    else:
        latt = np.array(lattice, dtype="double", order="C")

    _set_no_error()
    msg_type_list = _spglib.magnetic_spacegroup_type_from_symmetry(
        rots, trans, timerev, latt, symprec
    )

    if msg_type_list is not None:
        msg_type = MagneticSpaceGroupType(
            uni_number=msg_type_list[0],
            litvin_number=msg_type_list[1],
            bns_number=msg_type_list[2].strip(),
            og_number=msg_type_list[3].strip(),
            number=msg_type_list[4],
            type=msg_type_list[5],
        )
        return msg_type
    else:
        _set_error_message()
        return None


def get_pointgroup(rotations) -> tuple[str, int, np.ndarray]:
    """Return point group in international table symbol and number.

    The symbols are mapped to the numbers as follows:
    1   "1    "
    2   "-1   "
    3   "2    "
    4   "m    "
    5   "2/m  "
    6   "222  "
    7   "mm2  "
    8   "mmm  "
    9   "4    "
    10  "-4   "
    11  "4/m  "
    12  "422  "
    13  "4mm  "
    14  "-42m "
    15  "4/mmm"
    16  "3    "
    17  "-3   "
    18  "32   "
    19  "3m   "
    20  "-3m  "
    21  "6    "
    22  "-6   "
    23  "6/m  "
    24  "622  "
    25  "6mm  "
    26  "-62m "
    27  "6/mmm"
    28  "23   "
    29  "m-3  "
    30  "432  "
    31  "-43m "
    32  "m-3m "

    """
    _set_no_error()

    # (symbol, pointgroup_number, transformation_matrix)
    pointgroup = _spglib.pointgroup(np.array(rotations, dtype="intc", order="C"))
    _set_error_message()
    return pointgroup


def standardize_cell(
    cell: Cell,
    to_primitive=False,
    no_idealize=False,
    symprec=1e-5,
    angle_tolerance=-1.0,
) -> Cell | None:
    """Return standardized cell. When the search failed, ``None`` is returned.

    Parameters
    ----------
    cell, symprec, angle_tolerance:
        See the docstring of get_symmetry.
    to_primitive : bool
        If True, the standardized primitive cell is created.
    no_idealize : bool
        If True, it is disabled to idealize lengths and angles of basis vectors
        and positions of atoms according to crystal symmetry.

    Returns
    -------
    The standardized unit cell or primitive cell is returned by a tuple of
    (lattice, positions, numbers). If it fails, None is returned.

    Notes
    -----
    .. versionadded:: 1.8

    Now :func:`refine_cell` and :func:`find_primitive` are shorthands of
    this method with combinations of these options.
    About the default choice of the setting, see the documentation of ``hall_number``
    argument of :func:`get_symmetry_dataset`. More detailed explanation is
    shown in the spglib (C-API) document.

    """
    _set_no_error()

    lattice, _positions, _numbers, _ = _expand_cell(cell)

    # Atomic positions have to be specified by scaled positions for spglib.
    num_atom = len(_positions)
    positions = np.zeros((num_atom * 4, 3), dtype="double", order="C")
    positions[:num_atom] = _positions
    numbers = np.zeros(num_atom * 4, dtype="intc")
    numbers[:num_atom] = _numbers
    num_atom_std = _spglib.standardize_cell(
        lattice,
        positions,
        numbers,
        num_atom,
        to_primitive * 1,
        no_idealize * 1,
        symprec,
        angle_tolerance,
    )

    if num_atom_std > 0:
        return (
            np.array(lattice.T, dtype="double", order="C"),
            np.array(positions[:num_atom_std], dtype="double", order="C"),
            np.array(numbers[:num_atom_std], dtype="intc"),
        )
    else:
        _set_error_message()
        return None


def refine_cell(cell: Cell, symprec=1e-5, angle_tolerance=-1.0) -> Cell | None:
    """Return refined cell. When the search failed, ``None`` is returned.

    The standardized unit cell is returned by a tuple of
    (lattice, positions, numbers).

    Notes
    -----
    .. versionchanged:: 1.8

    The detailed control of standardization of unit cell can be done using
    :func:`standardize_cell`.

    """
    _set_no_error()

    lattice, _positions, _numbers, _ = _expand_cell(cell)

    # Atomic positions have to be specified by scaled positions for spglib.
    num_atom = len(_positions)
    positions = np.zeros((num_atom * 4, 3), dtype="double", order="C")
    positions[:num_atom] = _positions
    numbers = np.zeros(num_atom * 4, dtype="intc")
    numbers[:num_atom] = _numbers
    num_atom_std = _spglib.refine_cell(
        lattice,
        positions,
        numbers,
        num_atom,
        symprec,
        angle_tolerance,
    )

    if num_atom_std > 0:
        return (
            np.array(lattice.T, dtype="double", order="C"),
            np.array(positions[:num_atom_std], dtype="double", order="C"),
            np.array(numbers[:num_atom_std], dtype="intc"),
        )
    else:
        _set_error_message()
        return None


def find_primitive(cell: Cell, symprec=1e-5, angle_tolerance=-1.0) -> Cell | None:
    """Primitive cell is searched in the input cell. If it fails, ``None`` is returned.

    The primitive cell is returned by a tuple of (lattice, positions, numbers).

    Notes
    -----
    .. versionchanged:: 1.8

    The detailed control of standardization of unit cell can be done using
    :func:`standardize_cell`.

    """
    _set_no_error()

    lattice, positions, numbers, _ = _expand_cell(cell)

    num_atom_prim = _spglib.primitive(
        lattice, positions, numbers, symprec, angle_tolerance
    )

    if num_atom_prim > 0:
        return (
            np.array(lattice.T, dtype="double", order="C"),
            np.array(positions[:num_atom_prim], dtype="double", order="C"),
            np.array(numbers[:num_atom_prim], dtype="intc"),
        )
    else:
        _set_error_message()
        return None


def get_symmetry_from_database(hall_number) -> dict | None:
    """Return symmetry operations corresponding to a Hall symbol. If fails, return None.

    Parameters
    ----------
    hall_number : int
        The Hall symbol is given by the serial number in between 1 and 530.
        The definition of ``hall_number`` is found at
        :ref:`dataset_spg_get_dataset_spacegroup_type`.

    Returns
    -------
    symmetry : dict
        - rotations
            Rotation parts of symmetry operations corresponding to ``hall_number``.
        - translations
            Translation parts of symmetry operations corresponding to ``hall_number``.

    """
    _set_no_error()

    rotations = np.zeros((192, 3, 3), dtype="intc")
    translations = np.zeros((192, 3), dtype="double")
    num_sym = _spglib.symmetry_from_database(rotations, translations, hall_number)

    if num_sym is None:
        _set_error_message()
        return None
    else:
        return {
            "rotations": np.array(rotations[:num_sym], dtype="intc", order="C"),
            "translations": np.array(translations[:num_sym], dtype="double", order="C"),
        }


def get_magnetic_symmetry_from_database(uni_number, hall_number=0) -> dict | None:
    """Return magnetic symmetry operations from UNI number between 1 and 1651.

    If fails, return None.

    Optionally alternative settings can be specified with Hall number.

    Parameters
    ----------
    uni_number : int
        UNI number between 1 and 1651.
    hall_number : int, optional
        The Hall symbol is given by the serial number in between 1 and 530.

    Returns
    -------
    symmetry : dict
        - 'rotations'
        - 'translations'
        - 'time_reversals'
            0 and 1 indicate ordinary and anti-time-reversal operations, respectively.

    Notes
    -----
    .. versionadded:: 2.0

    """
    _set_no_error()

    rotations = np.zeros((384, 3, 3), dtype="intc")
    translations = np.zeros((384, 3), dtype="double")
    time_reversals = np.zeros(384, dtype="intc")
    num_sym = _spglib.magnetic_symmetry_from_database(
        rotations,
        translations,
        time_reversals,
        uni_number,
        hall_number,
    )

    if num_sym is None:
        _set_error_message()
        return None
    else:
        return {
            "rotations": np.array(rotations[:num_sym], dtype="intc", order="C"),
            "translations": np.array(translations[:num_sym], dtype="double", order="C"),
            "time_reversals": np.array(
                time_reversals[:num_sym],
                dtype="intc",
                order="C",
            ),
        }


############
# k-points #
############
def get_grid_point_from_address(grid_address, mesh):
    """Return grid point index by translating grid address."""
    _set_no_error()

    return _spglib.grid_point_from_address(
        np.array(grid_address, dtype="intc"),
        np.array(mesh, dtype="intc"),
    )


def get_ir_reciprocal_mesh(
    mesh,
    cell,
    is_shift=None,
    is_time_reversal=True,
    symprec=1e-5,
    is_dense=False,
):
    """Return k-points mesh and k-point map to the irreducible k-points.

    The symmetry is searched from the input cell.

    Parameters
    ----------
    mesh : array_like
        Uniform sampling mesh numbers.
        dtype='intc', shape=(3,)
    cell : spglib cell tuple
        Crystal structure.
    is_shift : array_like, optional
        [0, 0, 0] gives Gamma center mesh and value 1 gives half mesh shift.
        Default is None which equals to [0, 0, 0].
        dtype='intc', shape=(3,)
    is_time_reversal : bool, optional
        Whether time reversal symmetry is included or not. Default is True.
    symprec : float, optional
        Symmetry tolerance in distance. Default is 1e-5.
    is_dense : bool, optional
        grid_mapping_table is returned with dtype='uintp' if True. Otherwise
        its dtype='intc'. Default is False.

    Returns
    -------
    grid_mapping_table : ndarray
        Grid point mapping table to ir-gird-points.
        dtype='intc' or 'uintp', shape=(prod(mesh),)
    grid_address : ndarray
        Address of all grid points.
        dtype='intc', shape=(prod(mesh), 3)

    """
    _set_no_error()

    lattice, positions, numbers, _ = _expand_cell(cell)
    if lattice is None:
        return None

    if is_dense:
        dtype = "uintp"
    else:
        dtype = "intc"
    grid_mapping_table = np.zeros(np.prod(mesh), dtype=dtype)
    grid_address = np.zeros((np.prod(mesh), 3), dtype="intc")
    if is_shift is None:
        is_shift = [0, 0, 0]
    if (
        _spglib.ir_reciprocal_mesh(
            grid_address,
            grid_mapping_table,
            np.array(mesh, dtype="intc"),
            np.array(is_shift, dtype="intc"),
            is_time_reversal * 1,
            lattice,
            positions,
            numbers,
            symprec,
        )
        > 0
    ):
        return grid_mapping_table, grid_address
    else:
        _set_error_message()
        return None


def get_stabilized_reciprocal_mesh(
    mesh,
    rotations,
    is_shift=None,
    is_time_reversal=True,
    qpoints=None,
    is_dense=False,
):
    """Return k-point map to the irreducible k-points and k-point grid points.

    The symmetry is searched from the input rotation matrices in real space.

    Parameters
    ----------
    mesh : array_like
        Uniform sampling mesh numbers.
        dtype='intc', shape=(3,)
    rotations : array_like
        Rotation matrices with respect to real space basis vectors.
        dtype='intc', shape=(rotations, 3)
    is_shift : array_like
        [0, 0, 0] gives Gamma center mesh and value 1 gives  half mesh shift.
        dtype='intc', shape=(3,)
    is_time_reversal : bool
        Time reversal symmetry is included or not.
    qpoints : array_like
        q-points used as stabilizer(s) given in reciprocal space with respect
        to reciprocal basis vectors.
        dtype='double', shape=(qpoints ,3) or (3,)
    is_dense : bool, optional
        grid_mapping_table is returned with dtype='uintp' if True. Otherwise
        its dtype='intc'. Default is False.

    Returns
    -------
    grid_mapping_table : ndarray
        Grid point mapping table to ir-gird-points.
        dtype='intc', shape=(prod(mesh),)
    grid_address : ndarray
        Address of all grid points. Each address is given by three unsigned
        integers.
        dtype='intc', shape=(prod(mesh), 3)

    """
    _set_no_error()

    if is_dense:
        dtype = "uintp"
    else:
        dtype = "intc"
    mapping_table = np.zeros(np.prod(mesh), dtype=dtype)
    grid_address = np.zeros((np.prod(mesh), 3), dtype="intc")
    if is_shift is None:
        is_shift = [0, 0, 0]
    if qpoints is None:
        qpoints = np.array([[0, 0, 0]], dtype="double", order="C")
    else:
        qpoints = np.array(qpoints, dtype="double", order="C")
        if qpoints.shape == (3,):
            qpoints = np.array([qpoints], dtype="double", order="C")

    if (
        _spglib.stabilized_reciprocal_mesh(
            grid_address,
            mapping_table,
            np.array(mesh, dtype="intc"),
            np.array(is_shift, dtype="intc"),
            is_time_reversal * 1,
            np.array(rotations, dtype="intc", order="C"),
            qpoints,
        )
        > 0
    ):
        return mapping_table, grid_address
    else:
        _set_error_message()
        return None


def get_grid_points_by_rotations(
    address_orig,
    reciprocal_rotations,
    mesh,
    is_shift=None,
    is_dense=False,
):
    """Return grid points obtained after rotating input grid address.

    Parameters
    ----------
    address_orig : array_like
        Grid point address to be rotated.
        dtype='intc', shape=(3,)
    reciprocal_rotations : array_like
        Rotation matrices {R} with respect to reciprocal basis vectors.
        Defined by q'=Rq.
        dtype='intc', shape=(rotations, 3, 3)
    mesh : array_like
        dtype='intc', shape=(3,)
    is_shift : array_like, optional
        With (1) or without (0) half grid shifts with respect to grid intervals
        sampled along reciprocal basis vectors. Default is None, which
        gives [0, 0, 0].
    is_dense : bool, optional
        rot_grid_points is returned with dtype='uintp' if True. Otherwise
        its dtype='intc'. Default is False.

    Returns
    -------
    rot_grid_points : ndarray
        Grid points obtained after rotating input grid address
        dtype='intc' or 'uintp', shape=(rotations,)

    """
    _set_no_error()

    if is_shift is None:
        _is_shift = np.zeros(3, dtype="intc")
    else:
        _is_shift = np.array(is_shift, dtype="intc")

    rot_grid_points = np.zeros(len(reciprocal_rotations), dtype="uintp")
    _spglib.grid_points_by_rotations(
        rot_grid_points,
        np.array(address_orig, dtype="intc"),
        np.array(reciprocal_rotations, dtype="intc", order="C"),
        np.array(mesh, dtype="intc"),
        _is_shift,
    )

    if is_dense:
        return rot_grid_points
    else:
        return np.array(rot_grid_points, dtype="intc")


def get_BZ_grid_points_by_rotations(
    address_orig,
    reciprocal_rotations,
    mesh,
    bz_map,
    is_shift=None,
    is_dense=False,
):
    """Return grid points obtained after rotating input grid address.

    Parameters
    ----------
    address_orig : array_like
        Grid point address to be rotated.
        dtype='intc', shape=(3,)
    reciprocal_rotations : array_like
        Rotation matrices {R} with respect to reciprocal basis vectors.
        Defined by q'=Rq.
        dtype='intc', shape=(rotations, 3, 3)
    mesh : array_like
        dtype='intc', shape=(3,)
    bz_map : array_like
        TODO
    is_shift : array_like, optional
        With (1) or without (0) half grid shifts with respect to grid intervals
        sampled along reciprocal basis vectors. Default is None, which
        gives [0, 0, 0].
    is_dense : bool, optional
        rot_grid_points is returned with dtype='uintp' if True. Otherwise
        its dtype='intc'. Default is False.

    Returns
    -------
    rot_grid_points : ndarray
        Grid points obtained after rotating input grid address
        dtype='intc' or 'uintp', shape=(rotations,)

    """
    _set_no_error()

    if is_shift is None:
        _is_shift = np.zeros(3, dtype="intc")
    else:
        _is_shift = np.array(is_shift, dtype="intc")

    if bz_map.dtype == "uintp" and bz_map.flags.c_contiguous:
        _bz_map = bz_map
    else:
        _bz_map = np.array(bz_map, dtype="uintp")

    rot_grid_points = np.zeros(len(reciprocal_rotations), dtype="uintp")
    _spglib.BZ_grid_points_by_rotations(
        rot_grid_points,
        np.array(address_orig, dtype="intc"),
        np.array(reciprocal_rotations, dtype="intc", order="C"),
        np.array(mesh, dtype="intc"),
        _is_shift,
        _bz_map,
    )

    if is_dense:
        return rot_grid_points
    else:
        return np.array(rot_grid_points, dtype="intc")


def relocate_BZ_grid_address(
    grid_address,
    mesh,
    reciprocal_lattice,  # column vectors
    is_shift=None,
    is_dense=False,
):
    """Grid addresses are relocated to be inside first Brillouin zone.

    Number of ir-grid-points inside Brillouin zone is returned.
    It is assumed that the following arrays have the shapes of
        bz_grid_address : (num_grid_points_in_FBZ, 3)
        bz_map (prod(mesh * 2), )

    Note that the shape of grid_address is (prod(mesh), 3) and the
    addresses in grid_address are arranged to be in parallelepiped
    made of reciprocal basis vectors. The addresses in bz_grid_address
    are inside the first Brillouin zone or on its surface. Each
    address in grid_address is mapped to one of those in
    bz_grid_address by a reciprocal lattice vector (including zero
    vector) with keeping element order. For those inside first
    Brillouin zone, the mapping is one-to-one. For those on the first
    Brillouin zone surface, more than one addresses in bz_grid_address
    that are equivalent by the reciprocal lattice translations are
    mapped to one address in grid_address. In this case, those grid
    points except for one of them are appended to the tail of this array,
    for which bz_grid_address has the following data storing:

    .. code-block::

      |------------------array size of bz_grid_address-------------------------|
      |--those equivalent to grid_address--|--those on surface except for one--|
      |-----array size of grid_address-----|

    Number of grid points stored in bz_grid_address is returned.
    bz_map is used to recover grid point index expanded to include BZ
    surface from grid address. The grid point indices are mapped to
    (mesh[0] * 2) x (mesh[1] * 2) x (mesh[2] * 2) space (bz_map).

    """
    _set_no_error()

    if is_shift is None:
        _is_shift = np.zeros(3, dtype="intc")
    else:
        _is_shift = np.array(is_shift, dtype="intc")
    bz_grid_address = np.zeros((np.prod(np.add(mesh, 1)), 3), dtype="intc")
    bz_map = np.zeros(np.prod(np.multiply(mesh, 2)), dtype="uintp")
    num_bz_ir = _spglib.BZ_grid_address(
        bz_grid_address,
        bz_map,
        grid_address,
        np.array(mesh, dtype="intc"),
        np.array(reciprocal_lattice, dtype="double", order="C"),
        _is_shift,
    )

    if is_dense:
        return bz_grid_address[:num_bz_ir], bz_map
    else:
        return bz_grid_address[:num_bz_ir], np.array(bz_map, dtype="intc")


def delaunay_reduce(lattice, eps=1e-5):
    r"""Run Delaunay reduction. When the search failed, `None` is returned.

    The transformation from original basis vectors
    :math:`( \mathbf{a} \; \mathbf{b} \; \mathbf{c} )`
    to final basis vectors :math:`( \mathbf{a}' \; \mathbf{b}' \; \mathbf{c}' )` is achieved by linear
    combination of basis vectors with integer coefficients without
    rotating coordinates. Therefore the transformation matrix is obtained
    by :math:`\mathbf{P} = ( \mathbf{a} \; \mathbf{b} \; \mathbf{c} ) ( \mathbf{a}' \; \mathbf{b}' \; \mathbf{c}' )^{-1}` and the matrix
    elements have to be almost integers.

    The algorithm is found in the international tables for crystallography volume A.

    Parameters
    ----------
    lattice: ndarray, (3, 3)
        Lattice parameters in the form of

        .. code-block::

            [[a_x, a_y, a_z],
                [b_x, b_y, b_z],
                [c_x, c_y, c_z]]

    eps: float
        Tolerance parameter, but unlike `symprec` the unit is not a length.
        Tolerance to check if volume is close to zero or not and
        if two basis vectors are orthogonal by the value of dot
        product being close to zero or not.

    Returns
    -------
    delaunay_lattice: ndarray, (3, 3)
        Reduced lattice parameters are given as a numpy 'double' array:

        .. code-block::

            [[a_x, a_y, a_z],
             [b_x, b_y, b_z],
             [c_x, c_y, c_z]]

    Notes
    -----
    .. versionadded:: 1.9.4

    """  # noqa: E501
    _set_no_error()

    delaunay_lattice = np.array(np.transpose(lattice), dtype="double", order="C")
    result = _spglib.delaunay_reduce(delaunay_lattice, float(eps))

    if result == 0:
        _set_error_message()
        return None
    else:
        return np.array(np.transpose(delaunay_lattice), dtype="double", order="C")


def niggli_reduce(lattice, eps=1e-5):
    r"""Run Niggli reduction. When the search failed, ``None`` is returned.

    The transformation from original basis vectors :math:`( \mathbf{a} \; \mathbf{b} \; \mathbf{c} )` to final basis vectors :math:`( \mathbf{a}' \; \mathbf{b}' \; \mathbf{c}' )` is achieved by linear
    combination of basis vectors with integer coefficients without
    rotating coordinates. Therefore the transformation matrix is obtained
    by :math:`\mathbf{P} = ( \mathbf{a} \; \mathbf{b} \; \mathbf{c} ) ( \mathbf{a}' \; \mathbf{b}' \; \mathbf{c}' )^{-1}` and the matrix
    elements have to be almost integers.

    The algorithm detail is found at https://atztogo.github.io/niggli/ and
    the references are there in.

    Parameters
    ----------
    lattice: ndarray
        Lattice parameters in the form of

        .. code-block::

            [[a_x, a_y, a_z],
            [b_x, b_y, b_z],
            [c_x, c_y, c_z]]

    eps: float
        Tolerance parameter, but unlike `symprec` the unit is not a length.
        This is used to check if difference of norms of two basis
        vectors is close to zero or not and if two basis vectors are
        orthogonal by the value of dot product being close to zero or
        not.
        The detail is shown at https://atztogo.github.io/niggli/.

    Returns
    -------
    niggli_lattice: ndarray, (3, 3)
        if the Niggli reduction succeeded:
            Reduced lattice parameters are given as a numpy 'double' array:

            .. code-block::

                [[a_x, a_y, a_z],
                [b_x, b_y, b_z],
                [c_x, c_y, c_z]]

        otherwise None is returned.

    Notes
    -----
    .. versionadded:: 1.9.4

    """  # noqa: E501
    _set_no_error()

    niggli_lattice = np.array(np.transpose(lattice), dtype="double", order="C")
    result = _spglib.niggli_reduce(niggli_lattice, float(eps))

    if result == 0:
        _set_error_message()
        return None
    else:
        return np.array(np.transpose(niggli_lattice), dtype="double", order="C")


def get_error_message():
    """Return error message why spglib failed.

    .. warning::
        This method is not thread safe, i.e., only safely usable
        when calling one spglib method per process.

    Notes
    -----
    .. versionadded:: 1.9.5

    """
    return spglib_error.message


def _expand_cell(
    cell: Cell,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray | None]:
    lattice = np.array(np.transpose(cell[0]), dtype="double", order="C")
    positions = np.array(cell[1], dtype="double", order="C")
    numbers = np.array(cell[2], dtype="intc")
    if len(cell) == 4:
        magmoms = np.array(cell[3], order="C", dtype="double")
    elif len(cell) == 3:
        magmoms = None
    else:
        raise TypeError("cell has to be a tuple of 3 or 4 elements.")

    # Sanity check
    if lattice.shape != (3, 3):
        raise TypeError("lattice has to be a (3, 3) array.")
    if not (positions.ndim == 2 and positions.shape[1] == 3):
        raise TypeError("positions has to be a (num_atoms, 3) array.")
    num_atoms = positions.shape[0]
    if numbers.ndim != 1:
        raise TypeError("numbers has to be a (num_atoms,) array.")
    if len(numbers) != num_atoms:
        raise TypeError("numbers has to have the same number of atoms as positions.")
    if magmoms is not None:
        if len(magmoms) != num_atoms:
            raise TypeError(
                "magmoms has to have the same number of atoms as positions."
            )
        if magmoms.ndim == 1:
            # collinear
            pass
        elif magmoms.ndim == 2:
            # non-collinear
            if magmoms.shape[1] != 3:
                raise TypeError(
                    "non-collinear magmoms has to be a (num_atoms, 3) array."
                )
        else:
            raise TypeError("magmoms has to be a 1D or 2D array.")

    return (lattice, positions, numbers, magmoms)


def _set_error_message():
    spglib_error.message = _spglib.error_message()


def _set_no_error():
    spglib_error.message = "no error"


def get_hall_number_from_symmetry(rotations, translations, symprec=1e-5) -> int | None:
    """Hall number is obtained from a set of symmetry operations. If fails, return None.

    .. deprecated:: 2.0
        Replaced by {func}`get_spacegroup_type_from_symmetry`.

    Return one of ``hall_number`` corresponding to a space-group type of the given
    set of symmetry operations. When multiple ``hall_number`` exist for the
    space-group type, the smallest one (the first description of the space-group
    type in International Tables for Crystallography) is chosen. The definition of
    ``hall_number`` is found at :ref:`dataset_spg_get_dataset_spacegroup_type` and
    the corresponding space-group-type information is obtained through
    {func}`get_spacegroup_type`.

    This is expected to work well for the set of symmetry operations whose
    distortion is small. The aim of making this feature is to find
    space-group-type for the set of symmetry operations given by the other
    source than spglib.

    Note that the definition of ``symprec`` is
    different from usual one, but is given in the fractional
    coordinates and so it should be small like ``1e-5``.
    """
    warnings.warn(
        "get_hall_number_from_symmetry() is deprecated. "
        "Use get_spacegroup_type_from_symmetry() instead.",
        DeprecationWarning,
    )

    r = np.array(rotations, dtype="intc", order="C")
    t = np.array(translations, dtype="double", order="C")
    hall_number = _spglib.hall_number_from_symmetry(r, t, symprec)
    return hall_number



import re

from load import get_msg_numbers, get_msg_table


def get_type_of_msg(hall_symbol):
    # awkward classification...
    if "'" not in hall_symbol:
        return 1  # type-I
    if " 1'" in hall_symbol:
        return 2  # type-II
    if len(re.findall(r" 1[a-z]+'", hall_symbol)) > 0:
        return 4  # type-IV
    return 3  # type-III


if __name__ == '__main__':
    """
    Create `msg_database.c:magnetic_spacegroup_types`
    """
    msg_numbers = get_msg_numbers()

    msg_table = get_msg_table()
    bns_to_mhall = {}
    for dct in msg_table.values():
        bns_to_mhall.update(dct)

    contents = []

    contents.append("static const MagneticSpacegroupType magnetic_spacegroup_types[] = {")
    contents.append('    {    0,    0,        "",            "",   0, 0 }, /* Error */')

    max_bns_number_length = 7
    max_og_number_length = 11

    type_counts = {t: 0 for t in [1, 2, 3, 4]}

    # Sorted by UNI number
    for i, (litvin_number, bns_number, og_number, uni_number) in enumerate(msg_numbers):
        number = int(bns_number.split('.')[0])

        # type of MSG
        hall_symbol = bns_to_mhall[bns_number]
        type_of_msg = get_type_of_msg(hall_symbol)
        type_counts[type_of_msg] += 1

        assert len(bns_number) <= max_bns_number_length
        assert len(og_number) <= max_og_number_length
        quoted_bns_number = f'"{bns_number}"'
        quoted_og_number = f'"{og_number}"'
        contents.append(f'    {{ {uni_number:4d}, {litvin_number:4d}, {quoted_bns_number:>9s}, {quoted_og_number:>13s}, {number:3d}, {type_of_msg:1d} }}, /* {i+1:4d} */')
        assert uni_number == i + 1

    # Sanity check
    print(type_counts)
    assert type_counts[1] == 230
    assert type_counts[2] == 230
    assert type_counts[3] == 674
    assert type_counts[4] == 517

    contents.append("};")

    for line in contents:
        print(line)



import os
import unittest

import numpy as np
from spglib import find_primitive, get_symmetry_dataset
from vasp import read_vasp

data_dir = os.path.dirname(os.path.abspath(__file__))
dirnames = ("trigonal",)
rhomb_numbers = (146, 148, 155, 160, 161, 166, 167)
tmat = [
    [0.6666666666666666, -0.3333333333333333, -0.3333333333333333],
    [0.3333333333333333, 0.3333333333333333, -0.6666666666666666],
    [0.3333333333333333, 0.3333333333333333, 0.3333333333333333],
]


class TestRhombSettingHR(unittest.TestCase):
    def setUp(self):
        """Extract filename of rhombohedral cell."""
        self._filenames = []
        for d in dirnames:
            dirname = os.path.join(data_dir, "data", d)
            filenames = []
            trigo_filenames = os.listdir(dirname)
            for number in rhomb_numbers:
                filenames += [
                    fname for fname in trigo_filenames if str(number) in fname
                ]
            self._filenames += [os.path.join(dirname, fname) for fname in filenames]

    def tearDown(self):
        pass

    def test_rhomb_prim_agreement_over_settings(self):
        for fname in self._filenames:
            cell = read_vasp(fname)
            symprec = 1e-5
            dataset_H = get_symmetry_dataset(cell, symprec=symprec)
            hall_number_R = dataset_H.hall_number + 1
            dataset_R = get_symmetry_dataset(
                cell,
                hall_number=hall_number_R,
                symprec=symprec,
            )
            plat, _, _ = find_primitive(cell)
            plat_H = np.dot(dataset_H.std_lattice.T, tmat).T
            plat_R = dataset_R.std_lattice
            np.testing.assert_allclose(plat, plat_H, atol=1e-5, err_msg="%s" % fname)
            np.testing.assert_allclose(plat_R, plat_H, atol=1e-5, err_msg="%s" % fname)
            np.testing.assert_allclose(plat_R, plat, atol=1e-5, err_msg="%s" % fname)


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestRhombSettingHR)
    unittest.TextTestRunner(verbosity=2).run(suite)
    # unittest.main()



from transform import  get_standard_hall_number
from load import get_msg_numbers, get_spg_table, get_msg_table


if __name__ == '__main__':
    """
    Create
        magnetic_spacegroup_uni_mapping[uni_number][2] = {number of corresponding Hall numbers, the smallest Hall number}
    and
        magnetic_spacegroup_hall_mapping[hall_number][2] = {
            smallest UNI number corresponding to hall_number (in BNS setting),
            largest UNI number corresponding to hall_number (in BNS setting),
        }
    """
    # mapping: bns_number -> uni_number
    msg_numbers = get_msg_numbers()
    mapping = {}
    for _, bns_number, _, uni_number in msg_numbers:
        mapping[bns_number] = uni_number

    spg_table = get_spg_table()
    msg_table = get_msg_table()
    uni_to_hall = {uni_number: [] for uni_number in range(1, 1651 + 1)}
    hall_to_uni = {hall_number: [] for hall_number in range(1, 530 + 1)}

    for hall_number in range(1, 530 + 1):
        number = spg_table[hall_number]['number']
        hall_number_std = get_standard_hall_number(number)

        for bns_number in msg_table[hall_number_std].keys():
            uni_number = mapping[bns_number]
            uni_to_hall[uni_number].append(hall_number)
            hall_to_uni[hall_number].append(uni_number)

    contents = []
    contents.append("static const int magnetic_spacegroup_uni_mapping[][2] = {")
    contents.append("    {}, /* dummy */")
    for uni_number, hall_numbers in uni_to_hall.items():
        size = len(hall_numbers)
        first_hall_number = min(hall_numbers)
        contents.append(f'    {{ {size:2d}, {first_hall_number:3d} }}, /* {uni_number} */')
    contents.append("};")
    contents.append('')

    contents.append("static const int magnetic_spacegroup_hall_mapping[][2] = {")
    contents.append("    {}, /* dummy */")
    for hall_number, uni_numbers in hall_to_uni.items():
        min_uni_number = min(uni_numbers)
        max_uni_number = max(uni_numbers)
        contents.append(f'    {{ {min_uni_number:4d}, {max_uni_number:4d} }}, /* {hall_number:3d} */')
    contents.append("};")

    for line in contents:
        print(line)



"""Test of get_hall_number_from_symmetry."""

from __future__ import annotations

import os
from pathlib import Path
from typing import Callable

import pytest
from spglib import get_hall_number_from_symmetry, get_symmetry_dataset

data_dir = os.path.dirname(os.path.abspath(__file__))


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_get_hall_number_from_symmetry(all_filenames: list[Path], read_vasp: Callable):
    """Test get_hall_number_from_symmetry."""
    for fname in all_filenames:
        cell = read_vasp(fname)
        if "distorted" in str(fname):
            dataset = get_symmetry_dataset(cell, symprec=1e-1)
            hall_number = get_hall_number_from_symmetry(
                dataset.rotations,
                dataset.translations,
                symprec=1e-1,
            )
            if hall_number != dataset.hall_number:
                print("%d != %d in %s" % (hall_number, dataset.hall_number, fname))
                ref_cell = (
                    dataset.std_lattice,
                    dataset.std_positions,
                    dataset.std_types,
                )
                dataset = get_symmetry_dataset(ref_cell, symprec=1e-5)
                hall_number = get_hall_number_from_symmetry(
                    dataset.rotations,
                    dataset.translations,
                    symprec=1e-5,
                )
                print(
                    "Using refined cell: %d, %d in %s"
                    % (hall_number, dataset.hall_number, fname),
                )
        else:
            dataset = get_symmetry_dataset(cell, symprec=1e-5)
            hall_number = get_hall_number_from_symmetry(
                dataset.rotations,
                dataset.translations,
                symprec=1e-5,
            )
        assert hall_number == dataset.hall_number, "%d != %d in %s" % (
            hall_number,
            dataset.hall_number,
            fname,
        )



#!/usr/bin/env python

import sys

def read_spg_csv(filename="spg.csv"):
    spg_db = []
    for line in open(filename):
        data = line.split(',')
        spg_db.append(data)
    return spg_db

def extract_essense(spg_db):
    essence = []
    for x in spg_db:
        vals = []
        # IT number
        vals.append(int(x[4]))
        # SF
        vals.append(x[5])
        # Hall
        vals.append(x[6])
        # HM short all
        vals.append(x[7])
        # HM long
        vals.append(x[8])
        # HM short
        vals.append(x[7].split('=')[0].replace(' ',''))
        # Setting
        vals.append(x[2])
        # Centering
        vals.append(get_centering(x[8][0], ("%-9s" % x[2])[0]))
        essence.append(vals)
    return essence

def get_bravais(n, s):
    if 1 <= n and n <= 2:
        return 'triclinic'
    if 3 <= n and n <= 15:
        return 'monoclinic'
    if 16 <= n and n <= 74:
        return 'orthorhombic'
    if 75 <= n and n <= 142:
        return 'tetragonal'
    if 143 <= n and n <= 167:
        if is_rhombo( s ):
            return 'rhombohedral'
        else:
            return 'trigonal'
    if 168 <= n and n <= 194:
        return 'hexagonal'
    if 195 <= n and n <= 230:
        return 'cubic'

def get_pointgroup_number(spg_num):
    spg_num_head = [1, 2, 3, 6, 10, 16, 25, 47, 75, 81,
                    83, 89, 99, 111, 123, 143, 147, 149, 156, 162,
                    168, 174, 175, 177, 183, 187, 191, 195, 200, 207,
                    215, 221]
    pt_num = 0
    for n in spg_num_head:
        if spg_num >= n:
            pt_num += 1

    return pt_num

def get_centering(centering, setting):
    if centering == 'I':
        return "BODY"
    elif centering == 'F':
        return "FACE"
    elif centering == 'A':
        return "A_FACE"
    elif centering == 'C':
        return "C_FACE"
    elif centering == 'R':
        if setting == 'H':
            return "R_CENTER"
        else:
            return "PRIMITIVE"
    else:
        return "PRIMITIVE"

def is_rhombo( s ):
    if s=='R':
        return True

maxlen = 0

bravais = {'triclinic': 'TRICLI',
           'monoclinic': 'MONOCLI',
           'orthorhombic': 'ORTHO',
           'tetragonal': 'TETRA',
           'trigonal': 'TRIGO',
           'rhombohedral': 'RHOMB',
           'hexagonal': 'HEXA',
           'cubic': 'CUBIC'}

print "static const SpacegroupType spacegroup_types[] = {"

print "  {%3d, \"%-6s\", \"%-16s\", \"%-31s\", \"%-19s\", \"%-10s\", \"%-5s\", %15s, %2d }, /* %3d */" % (0, "", "", "", "", "", "", "CENTERING_ERROR", 0, 0)

for i, x in enumerate(extract_essense(read_spg_csv(sys.argv[1]))):
    if len(x[6]) > maxlen:
        maxlen = len(x[6])

    pt_num = get_pointgroup_number(x[0])

    print "  {%3d, \"%-6s\", \"%-16s\", \"%-31s\", \"%-19s\", \"%-10s\", \"%-5s\", %15s, %2d }, /* %3d */" % (x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7], pt_num, i + 1)
print "};"



#!/usr/bin/env python

# This will (hopefully) be the code to extract symmetry operations
# from Hall symbols

import numpy as np

lattice_symbols = {
    'P': [[0, 0, 0]],
    'A': [[0, 0, 0], [0, 1./2, 1./2]],
    'B': [[0, 0, 0], [1./2, 0, 1./2]],
    'C': [[0, 0, 0], [1./2, 1./2, 0]],
    'I': [[0, 0, 0], [1./2, 1./2, 1./2]],
    'R': [[0, 0, 0], [2./3, 1./3, 1./3], [1./3, 2./3, 2./3]],
    'H': [[0, 0, 0], [2./3, 1./3, 0], [1./3, 2./3, 0]],
    'F': [[0, 0, 0], [0, 1./2, 1./2], [1./2, 0, 1./2], [1./2, 1./2, 0]],
    'p': [[0, 0, 0]],
    'c': [[0, 0, 0], [1./2, 1./2, 0]]
}

rotation_matrices = {
    '1x': [[1, 0, 0],
           [0, 1, 0],
           [0, 0, 1]],
    '1y': [[1, 0, 0],
           [0, 1, 0],
           [0, 0, 1]],
    '1z': [[1, 0, 0],
           [0, 1, 0],
           [0, 0, 1]],
    '2x': [[1, 0, 0],
           [0, -1, 0],
           [0, 0, -1]],
    '2y': [[-1, 0, 0],
           [0, 1, 0],
           [0, 0, -1]],
    '2z': [[-1, 0, 0],
           [0, -1, 0],
           [0, 0, 1]],
    '3x': [[1, 0, 0],
           [0, 0, -1],
           [0, 1, -1]],
    '3y': [[-1, 0, 1],
           [0, 1, 0],
           [-1, 0, 0]],
    '3z': [[0, -1, 0],
           [1, -1, 0],
           [0, 0, 1]],
    '4x': [[1, 0, 0],
           [0, 0, -1],
           [0, 1, 0]],
    '4y': [[0, 0, 1],
           [0, 1, 0],
           [-1, 0, 0]],
    '4z': [[0, -1, 0],
           [1, 0, 0],
           [0, 0, 1]],
    '6x': [[1, 0, 0],
           [0, 1, -1],
           [0, 1, 0]],
    '6y': [[0, 0, 1],
           [0, 1, 0],
           [-1, 0, 1]],
    '6z': [[1, -1, 0],
           [1, 0, 0],
           [0, 0, 1]],
    '2px': [[-1, 0, 0],     # b-c
            [0, 0, -1],
            [0, -1, 0]],
    '2ppx': [[-1, 0, 0],    # b+c
             [0, 0, 1],
             [0, 1, 0]],
    '2py': [[0, 0, -1],     # a-c
            [0, -1, 0],
            [-1, 0, 0]],
    '2ppy': [[0, 0, 1],    # a+c
             [0, -1, 0],
             [1, 0, 0]],
    '2pz': [[0, -1, 0],     # a-b
            [-1, 0, 0],
            [0, 0, -1]],
    '2ppz': [[0, 1, 0],    # a+b
             [1, 0, 0],
             [0, 0, -1]],
    '3*': [[0, 0, 1],     # a+b+c
           [1, 0, 0],
           [0, 1, 0]]
}

translations = {
    'a': [1./2, 0, 0],
    'b': [0, 1./2, 0],
    'c': [0, 0, 1./2],
    'n': [1./2, 1./2, 1./2],
    'u': [1./4, 0, 0],
    'v': [0, 1./4, 0],
    'w': [0, 0, 1./4],
    'd': [1./4, 1./4, 1./4]
}


def read_spg_csv(filename="spg.csv"):
    hall_symbols = []
    for line in open(filename):
        data = line.split(',')
        hall_symbols.append([data[6], data[4]])
    return hall_symbols


class HallSymbol:
    def __init__(self, hall_symbol):
        self.hall_symbol = hall_symbol.split()
        self._decompose()
        self._full_operations()

    def get_LNV(self):
        return self.L, self.N, self.V

    def get_operations(self):
        return self.G_R, self.G_T

    def _full_operations(self):
        gens_R, gens_T = self._generators()

        E = np.array(rotation_matrices['1x'])
        T0 = np.zeros(3, dtype=float)
        if self.L[0] == '-':
            G_R = [E, -E]
            G_T = [T0, T0]
        else:
            G_R = [E]
            G_T = [T0]

        for r, t in zip(gens_R, gens_T):
            G2_R, G2_T = self._get_group(r, t)
            G_R, G_T = self._multiply_groups(G_R, G_T, G2_R, G2_T)

        if self.V is not None:
            G_T = self._change_of_basis(G_R, G_T)

        G_R_with_centres = []
        G_T_with_centred = []
        for t in lattice_symbols[self.L[-1]]:
            self._lattice_translation(G_R_with_centres,
                                      G_T_with_centred,
                                      G_R, G_T, t)

        self.G_R = np.array(G_R_with_centres)
        self.G_T = np.array(G_T_with_centred)

        # Make sure the first operation has no rotation.
        assert (self.G_R[0] == rotation_matrices['1x']).all()

        # In Hall numbers 212, 213, 214, the first operation has non-zero
        # translation. This translation is subtracted from all operations.
        self.G_T -= self.G_T[0]
        self.G_T -= np.rint(self.G_T)
        cond = self.G_T < -1e-3
        self.G_T[cond] += 1

    def _change_of_basis(self, G_R, G_T):
        G_T_new = []
        v = self.V.astype(float) / 12
        for r, t in zip(G_R, G_T):
            G_T_new.append(-np.dot(r, v) + t + v)
        return G_T_new

    def _lattice_translation(self, G_R, G_T, G_R0, G_T0, translation):
        for r, t in zip(G_R0, G_T0):
            G_R.append(r.copy())
            t_new = t + translation
            G_T.append(t_new)

    def _multiply_groups(self, G1_R, G1_T, G2_R, G2_T):  # G2xG1
        G_R = []
        G_T = []
        for r1, t1 in zip(G2_R, G2_T):
            for r2, t2 in zip(G1_R, G1_T):
                G_R.append(np.dot(r1, r2))
                G_T.append(np.dot(r1, t2) + t1)

        return G_R, G_T

    def _get_group(self, r, t):
        G_R = [r, ]
        G_T = [t, ]
        while not (G_R[-1] == rotation_matrices['1x']).all():
            _r = np.dot(G_R[-1], r)
            _t = np.dot(G_R[-1], t) + G_T[-1]
            G_R.append(_r)
            G_T.append(_t)

        # Bring identity in front
        _r = G_R.pop()
        _t = G_T.pop()
        G_R.insert(0, _r)
        G_T.insert(0, _t)

        return G_R, G_T

    # def _get_group(self, r, t):
    #     G_R, G_T = self._get_group_recursive([np.array(r)], [np.array(t)])
    #     r = G_R.pop()
    #     t = G_T.pop()
    #     G_R.insert(0, r)
    #     G_T.insert(0, t)
    #     return G_R, G_T

    # def _get_group_recursive(self, G_R, G_T):
    #     if not (G_R[-1] == rotation_matrices['1x']).all():
    #         r = np.dot(G_R[-1], G_R[0])
    #         t = np.dot(G_R[-1], G_T[0]) + G_T[-1]
    #         G_R.append(r)
    #         G_T.append(t)
    #         self._get_group_recursive(G_R, G_T)
    #     return G_R, G_T

    def _generators(self):
        R = []
        T = []
        for N in self.N:
            rot = np.array(rotation_matrices[N[1] + N[2]])
            if N[0] == '-':
                rot = -rot
            R.append(rot)

            trans = np.zeros(3, dtype=float)
            if N[3] is not None:
                for t in N[3]:
                    if t in ('1', '2', '3', '4', '5'):
                        trans_screw = float(t) / int(N[1])
                        if N[2] == 'x':
                            trans[0] += trans_screw
                        elif N[2] == 'y':
                            trans[1] += trans_screw
                        elif N[2] == 'z':
                            trans[2] += trans_screw
                        else:
                            raise
                    else:
                        trans += np.array(translations[t])
            T.append(trans)

        return np.array(R, dtype=int), np.array(T, dtype=float)

    def _rotation_matrix(self, str):
        pass

    # Decompose Hall symbol
    # The following methods are used by _decompose().
    def _decompose(self):
        L = self.hall_symbol.pop(0)
        N = []
        V = None
        precededN = 0
        for i, ms in enumerate(self.hall_symbol):
            if ms[0] == '(':
                V = self._change_of_basis_symbol(self.hall_symbol[i + 2])
                break
            else:
                N.append(self._matrix_symbol(ms, i, precededN))
                precededN = int(N[-1][1][0])

        self.L = L
        self.N = N
        self.V = V

    def _matrix_symbol(self, N, i, precededN):
        if N[0] == '-':
            improper = '-'
            N = N[1:]
        else:
            improper = None

        N, R, A = self._rotation(N, i, precededN)

        if len(N) > 0:
            T = self._translation(N)
        else:
            T = None

        return [improper, R, A, T]

    def _rotation(self, N, i, precededN):
        A = None
        if N[0] == '2':
            if len(N) > 1:  # 2"
                if N[1] == '=':
                    R = '2pp'
                    A = 'z'
                    N = N[2:]
            if i == 1 and A is None:
                if precededN == 2 or precededN == 4:  # 2x
                    R = '2'
                    A = 'x'
                    N = N[1:]
                elif precededN == 3 or precededN == 6:  # 2'
                    R = '2p'
                    A = 'z'
                    N = N[1:]
        elif N[0] == '3':     # 3*
            if i == 2:
                R = '3'
                A = '*'
                N = N[1:]
            elif len(N) > 1:
                if N[1] == '*':
                    R = '3'
                    A = '*'
                    N = N[2:]

        if A is None:
            R = N[0]
            N = N[1:]
            if len(N) > 0 and i == 0:
                N, A = self._principal_axis(N)
            else:
                A = 'z'

        return N, R, A

    def _principal_axis(self, N):
        if N[0] == 'x':
            return N[1:], 'x'
        if N[0] == 'y':
            return N[1:], 'y'
        return N, 'z'

    def _translation(self, N):
        T = []
        for i in range(len(N)):
            T.append(N[i])
        return T

    def _change_of_basis_symbol(self, V):
        return np.array([0, 0, int(V[:-1])])


def dump_operations(filename):
    hall_symbols = read_spg_csv(filename)
    count = 0
    print("  0       ,  /* dummy */")
    for i in range(len(hall_symbols)):
        hs = HallSymbol(hall_symbols[i][0])
        G_R, G_T = hs.get_operations()
        for j, (r, t) in enumerate(zip(G_R, G_T)):
            count += 1
            total = encode_symmetry(r, t)
            r_enc_dec, t_enc_dec = decode_symmetry(total)
            assert np.allclose(r_enc_dec, r.reshape(-1).tolist())
            assert np.allclose(t_enc_dec, t * 12)

            text = "  %-8d," % (total)
            text += " /* %4d (%3d) [" % (count, i + 1)
            text += "%2d," * 9 % tuple(r_enc_dec)
            text += "%2d,%2d,%2d] */" % tuple(t_enc_dec)
            print(text)


def dump_operations_old(filename):
    hall_symbols = read_spg_csv(filename)
    count = 0
    for i in range(530):
        hs = HallSymbol(hall_symbols[i][0])
        G_R, G_T = hs.get_operations()
        for j, (r, t) in enumerate(zip(G_R, G_T)):
            count += 1
            text = "{%3d," % (i + 1)
            text += "%2d,%2d,%2d,%2d,%2d,%2d,%2d,%2d,%2d," % tuple(r.ravel())
            text += "%2d,%2d,%2d" % tuple((t * 12 + 0.1).astype(int))
            text += "}, /* %4d */" % count
            print(text)


# Ternary numerical system
def encode_symmetry(r, t):
    r_encode = encode_rotation(r)
    t_encode = encode_trans(t)
    total = t_encode * 3 ** 9 + r_encode
    return total


def encode_rotation(r):
    r_sum = 0
    for i, x in enumerate(r.ravel()):
        r_sum += (x + 1) * 3**(8 - i)
    return r_sum


def encode_trans(t):
    x = np.rint(t * 12).astype(int)
    t_encode = x[0] * 144 + x[1] * 12 + x[2]
    return t_encode


def decode_symmetry(c):
    r_encoded = c % (3 ** 9)
    t_encoded = c // (3 ** 9)
    r_aligned = decode_rotation(r_encoded)  # (9, )
    t_multiplied = decode_trans(t_encoded)  # (3, )
    return r_aligned, t_multiplied


def decode_rotation(c):
    r = []
    for i in range(8, -1, -1):
        r.append((c % (3**(i+1))) // (3**i) - 1)
    return np.array(r)


def decode_trans(c):
    return c // 144, (c % 144) // 12, (c % 12)


def get_reference_to_operations(filename):
    hall_symbols = read_spg_csv(filename)
    count = 0
    for i in range(len(hall_symbols)):
        hs = HallSymbol(hall_symbols[i][0])
        G_R, G_T = hs.get_operations()
        print("  {%4d,%5d}, /* %3d */ " % (len(G_R), count + 1, i + 1))
        count += len(G_R)


def watch_hs(filename, number):
    print("  {   0,    0}, /*   0 */")
    num = number - 1
    hall_symbols = read_spg_csv(filename)
    hs = HallSymbol(hall_symbols[num][0])
    for char, vals in zip(('L', 'N', 'V'), hs.get_LNV()):
        print("%s: %s" % (char, vals))
    G_R, G_T = hs.get_operations()
    print(number, ":", hall_symbols[num][0], "(", len(G_R), ")")
    for i, (r, t) in enumerate(zip(G_R, G_T)):
        print("-----", i + 1, "-----")
        print(r, t)


if __name__ == '__main__':
    """

    Usage
    -----

    To watch symmetry operations of a Hall symbol,

    % python hall2operations.py --hs spg.csv 213

    To dump symmetry operations of all Hall symbols that are copied to
    spglb_database.c,

    % python hall2operations.py --dump spg.csv

    To dump address of symmetry operation data that is copied to
    spglb_database.c,

    % python hall2operations.py --reference spg.csv

    """

    from optparse import OptionParser
    parser = OptionParser()
    parser.set_defaults(watch_hs=False,
                        dump_operations=False,
                        shift=None,
                        origin=None)
    parser.add_option("--hs", dest="watch_hs",
                      action="store_true",
                      help="spg.csv [spg NUM]")
    parser.add_option("--dump", dest="is_dump",
                      action="store_true")
    parser.add_option("--reference", dest="is_reference",
                      action="store_true")

    (options, args) = parser.parse_args()

    if options.is_dump:
        dump_operations(args[0])
    if options.is_reference:
        get_reference_to_operations(args[0])
    elif options.watch_hs:
        watch_hs(args[0], int(args[1]))



import os
import unittest

import numpy as np
import yaml
from spglib import (
    MagneticSpaceGroupType,
    find_primitive,
    get_magnetic_spacegroup_type,
    get_magnetic_symmetry_from_database,
    get_pointgroup,
    get_spacegroup,
    get_spacegroup_type,
    get_symmetry_dataset,
    standardize_cell,
)
from vasp import read_vasp

data_dir = os.path.dirname(os.path.abspath(__file__))

dirnames = (
    "cubic",
    "hexagonal",
    "monoclinic",
    "orthorhombic",
    "tetragonal",
    "triclinic",
    "trigonal",
    "distorted",
    "virtual_structure",
)

# fmt: off
spg_to_hall = [
    1,   2,   3,   6,   9,   18,  21,  30,  39,  57,  # noqa: E241
    60,  63,  72,  81,  90,  108, 109, 112, 115, 116,  # noqa: E241
    119, 122, 123, 124, 125, 128, 134, 137, 143, 149,
    155, 161, 164, 170, 173, 176, 182, 185, 191, 197,
    203, 209, 212, 215, 218, 221, 227, 228, 230, 233,
    239, 245, 251, 257, 263, 266, 269, 275, 278, 284,
    290, 292, 298, 304, 310, 313, 316, 322, 334, 335,
    337, 338, 341, 343, 349, 350, 351, 352, 353, 354,
    355, 356, 357, 358, 359, 361, 363, 364, 366, 367,
    368, 369, 370, 371, 372, 373, 374, 375, 376, 377,
    378, 379, 380, 381, 382, 383, 384, 385, 386, 387,
    388, 389, 390, 391, 392, 393, 394, 395, 396, 397,
    398, 399, 400, 401, 402, 404, 406, 407, 408, 410,
    412, 413, 414, 416, 418, 419, 420, 422, 424, 425,
    426, 428, 430, 431, 432, 433, 435, 436, 438, 439,
    440, 441, 442, 443, 444, 446, 447, 448, 449, 450,
    452, 454, 455, 456, 457, 458, 460, 462, 463, 464,
    465, 466, 467, 468, 469, 470, 471, 472, 473, 474,
    475, 476, 477, 478, 479, 480, 481, 482, 483, 484,
    485, 486, 487, 488, 489, 490, 491, 492, 493, 494,
    495, 497, 498, 500, 501, 502, 503, 504, 505, 506,
    507, 508, 509, 510, 511, 512, 513, 514, 515, 516,
    517, 518, 520, 521, 523, 524, 525, 527, 529, 530, 531]
# fmt: on


class TestSpglib(unittest.TestCase):
    def setUp(self):
        self._filenames = []
        self._ref_filenames = []
        self._spgnum_ref = []
        for d in dirnames:
            dirname = os.path.join(data_dir, "data", d)
            refdirname = os.path.join(data_dir, "ref", d)
            filenames = os.listdir(dirname)
            self._spgnum_ref += [int(fname.split("-")[1]) for fname in filenames]
            self._filenames += [os.path.join(dirname, fname) for fname in filenames]
            self._ref_filenames += [
                os.path.join(refdirname, fname + "-ref") for fname in filenames
            ]

    def _create_symref(self):
        pass

    def tearDown(self):
        pass

    def test_get_symmetry_dataset(self):
        for fname, spgnum, reffname in zip(
            self._filenames,
            self._spgnum_ref,
            self._ref_filenames,
        ):
            cell = read_vasp(fname)

            if "distorted" in fname:
                symprec = 1e-1
            else:
                symprec = 1e-5
            dataset = get_symmetry_dataset(cell, symprec=symprec)
            self.assertEqual(dataset.number, spgnum, msg=("%s" % fname))

            for i in range(spg_to_hall[spgnum - 1], spg_to_hall[spgnum]):
                dataset = get_symmetry_dataset(cell, hall_number=i, symprec=symprec)
                self.assertEqual(dataset.hall_number, i, msg=("%s" % fname))
                spg_type = get_spacegroup_type(dataset.hall_number)
                self.assertEqual(
                    dataset.international,
                    spg_type.international_short,
                    msg=("%s" % fname),
                )
                self.assertEqual(
                    dataset.hall,
                    spg_type.hall_symbol,
                    msg=("%s" % fname),
                )
                self.assertEqual(
                    dataset.choice,
                    spg_type.choice,
                    msg=("%s" % fname),
                )
                self.assertEqual(
                    dataset.pointgroup,
                    spg_type.pointgroup_international,
                    msg=("%s" % fname),
                )

            wyckoffs = dataset.wyckoffs
            with open(reffname) as f:
                wyckoffs_ref = yaml.load(f, Loader=yaml.FullLoader)["wyckoffs"]
            for w, w_ref in zip(wyckoffs, wyckoffs_ref):
                self.assertEqual(w, w_ref, msg=("%s" % fname))

            # This is for writing out detailed symmetry info into files.
            # Now it is only for Wyckoff positions.
            # with open(reffname, 'w') as f:
            #     f.write("wyckoffs:\n")
            #     for w in dataset['wyckoffs']:
            #         f.write("- \"%s\"\n" % w)

    def test_standardize_cell_and_pointgroup(self):
        for fname, spgnum in zip(self._filenames, self._spgnum_ref):
            cell = read_vasp(fname)
            if "distorted" in fname:
                symprec = 1e-1
            else:
                symprec = 1e-5

            std_cell = standardize_cell(
                cell,
                to_primitive=False,
                no_idealize=True,
                symprec=symprec,
            )
            dataset = get_symmetry_dataset(std_cell, symprec=symprec)
            self.assertEqual(dataset.number, spgnum, msg=("%s" % fname))

            # The test for point group has to be done after standardization.
            ptg_symbol, _, _ = get_pointgroup(dataset.rotations)
            self.assertEqual(dataset.pointgroup, ptg_symbol, msg=("%s" % fname))

    def test_standardize_cell_from_primitive(self):
        for fname, spgnum in zip(self._filenames, self._spgnum_ref):
            cell = read_vasp(fname)
            if "distorted" in fname:
                symprec = 1e-1
            else:
                symprec = 1e-5

            prim_cell = standardize_cell(
                cell,
                to_primitive=True,
                no_idealize=True,
                symprec=symprec,
            )
            std_cell = standardize_cell(
                prim_cell,
                to_primitive=False,
                no_idealize=True,
                symprec=symprec,
            )
            dataset = get_symmetry_dataset(std_cell, symprec=symprec)
            self.assertEqual(dataset.number, spgnum, msg=("%s" % fname))

    def test_standardize_cell_to_primitive(self):
        for fname, spgnum in zip(self._filenames, self._spgnum_ref):
            cell = read_vasp(fname)
            if "distorted" in fname:
                symprec = 1e-1
            else:
                symprec = 1e-5

            prim_cell = standardize_cell(
                cell,
                to_primitive=True,
                no_idealize=True,
                symprec=symprec,
            )
            dataset = get_symmetry_dataset(prim_cell, symprec=symprec)
            self.assertEqual(dataset.number, spgnum, msg=("%s" % fname))

    def test_refine_cell(self):
        for fname, spgnum in zip(self._filenames, self._spgnum_ref):
            cell = read_vasp(fname)
            if "distorted" in fname:
                dataset_0 = get_symmetry_dataset(cell, symprec=1e-1)
            else:
                dataset_0 = get_symmetry_dataset(cell, symprec=1e-5)
            ref_cell_0 = (
                dataset_0.std_lattice,
                dataset_0.std_positions,
                dataset_0.std_types,
            )
            dataset_1 = get_symmetry_dataset(ref_cell_0, symprec=1e-5)
            # Check the same space group type is found.
            self.assertEqual(dataset_1.number, spgnum, msg=("%s" % fname))

            # Check if the same structure is obtained when applying
            # standardization again, i.e., examining non cycling behaviour.
            # Currently only for orthorhombic.
            if (
                "cubic" in fname
                or "hexagonal" in fname
                or "monoclinic" in fname
                or "orthorhombic" in fname
                or "tetragonal" in fname
                or "triclinic" in fname
                or "trigonal" in fname
                or "distorted" in fname
            ):
                ref_cell_1 = (
                    dataset_1.std_lattice,
                    dataset_1.std_positions,
                    dataset_1.std_types,
                )
                dataset_2 = get_symmetry_dataset(ref_cell_1, symprec=1e-5)
                np.testing.assert_equal(
                    dataset_1.std_types,
                    dataset_2.std_types,
                    err_msg="%s" % fname,
                )
                np.testing.assert_allclose(
                    dataset_1.std_lattice,
                    dataset_2.std_lattice,
                    atol=1e-5,
                    err_msg="%s" % fname,
                )
                diff = dataset_1.std_positions - dataset_2.std_positions
                diff -= np.rint(diff)
                np.testing.assert_allclose(diff, 0, atol=1e-5, err_msg="%s" % fname)

    def test_get_spacegroup(self):
        cell = (
            [
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0],
            ],
            [
                [0.0, 0.0, 0.0],
                [0.0, 0.5, 0.5],
                [0.5, 0.0, 0.5],
                [0.5, 0.5, 0.0],
            ],
            [0, 0, 0, 0],
        )
        assert get_spacegroup(cell) == "Fm-3m (225)"
        assert get_spacegroup(cell, symbol_type=1) == "Oh^5 (225)"

    def test_find_primitive(self):
        for fname in self._filenames:
            cell = read_vasp(fname)
            if "distorted" in fname:
                symprec = 1e-1
            else:
                symprec = 1e-5

            dataset = get_symmetry_dataset(cell, symprec=symprec)
            primitive = find_primitive(cell, symprec=symprec)

            spg_type = get_spacegroup_type(dataset.hall_number)
            c = spg_type.international_short[0]
            if c in ["A", "B", "C", "I"]:
                multiplicity = 2
            elif c == "F":
                multiplicity = 4
            elif c == "R":
                self.assertEqual(spg_type.choice, "H")
                if spg_type.choice == "H":
                    multiplicity = 3
                else:  # spg_type['choice'] == 'R'
                    multiplicity = 1
            else:
                multiplicity = 1
            self.assertEqual(
                len(dataset.std_types),
                len(primitive[2]) * multiplicity,
                msg=("multi: %d, %s" % (multiplicity, fname)),
            )

    def test_magnetic_spacegroup_type(self):
        # P 3 -2"
        actual1 = get_magnetic_spacegroup_type(1279)
        expect1 = MagneticSpaceGroupType(
            uni_number=1279,
            litvin_number=1279,
            bns_number="156.49",
            og_number="156.1.1279",
            number=156,
            type=1,
        )
        assert actual1 == expect1

        # -P 2 2ab 1'
        actual2 = get_magnetic_spacegroup_type(452)
        expect2 = MagneticSpaceGroupType(
            uni_number=452,
            litvin_number=442,
            bns_number="55.354",
            og_number="55.2.442",
            number=55,
            type=2,
        )
        assert actual2 == expect2

        # P 31 2 1c' (0 0 1)
        actual3 = get_magnetic_spacegroup_type(1262)
        expect3 = MagneticSpaceGroupType(
            uni_number=1262,
            litvin_number=1270,
            bns_number="151.32",
            og_number="153.4.1270",
            number=151,
            type=4,
        )
        assert actual3 == expect3

    def test_magnetic_symmetry_database(self):
        # UNI: R31'_c[R3] (1242), BNS: R_I3 (146.12)

        # Hexagonal axes: hall_number: 433
        data_h_actual = get_magnetic_symmetry_from_database(1242)
        for key in ["rotations", "translations", "time_reversals"]:
            assert len(data_h_actual[key]) == 18

        # Rhombohedral axes: hall_number: 434
        data_r_actual = get_magnetic_symmetry_from_database(1242, hall_number=434)
        data_r_expect = {
            "rotations": np.array(
                [
                    # x,y,z
                    [
                        [1, 0, 0],
                        [0, 1, 0],
                        [0, 0, 1],
                    ],
                    # y,z,x
                    [
                        [0, 0, 1],
                        [1, 0, 0],
                        [0, 1, 0],
                    ],
                    # y+1/2,z+1/2,x+1/2'
                    [
                        [0, 0, 1],
                        [1, 0, 0],
                        [0, 1, 0],
                    ],
                    # z,x,y
                    [
                        [0, 1, 0],
                        [0, 0, 1],
                        [1, 0, 0],
                    ],
                    # x+1/2,y+1/2,z+1/2'
                    [
                        [1, 0, 0],
                        [0, 1, 0],
                        [0, 0, 1],
                    ],
                    # z+1/2,x+1/2,y+1/2'
                    [
                        [0, 1, 0],
                        [0, 0, 1],
                        [1, 0, 0],
                    ],
                ],
                dtype=np.int32,
            ),
            "translations": np.array(
                [
                    [0, 0, 0],
                    [0, 0, 0],
                    [0.5, 0.5, 0.5],
                    [0, 0, 0],
                    [0.5, 0.5, 0.5],
                    [0.5, 0.5, 0.5],
                ],
            ),
            "time_reversals": np.array(
                [
                    [0, 0, 1, 0, 1, 1],
                ],
            ),
        }
        for key in ["rotations", "translations", "time_reversals"]:
            assert np.allclose(data_r_actual[key], data_r_expect[key])


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestSpglib)
    unittest.TextTestRunner(verbosity=2).run(suite)
    # unittest.main()



from pathlib import Path
import csv

from ruamel.yaml import YAML


def get_msg_numbers():
    all_datum = []
    with open(Path(__file__).resolve().parent / "msg_numbers.csv", 'r') as f:
        reader = csv.reader(f, delimiter=',')
        next(reader)  # skip header
        for row in reader:
            if len(row) == 0:
                break

            litvin_number, bns_number, og_number, uni_number = row
            all_datum.append((
                int(litvin_number),
                bns_number,
                og_number,
                int(uni_number),
            ))

    assert len(all_datum) == 1651
    return all_datum


def get_spg_table():
    all_datum = {}
    with open(Path("../spg.csv"), 'r') as f:
        reader = csv.reader(f, delimiter=',')
        for row in reader:
            hall_number, choice, number, hall_symbol = int(row[0]), row[2], int(row[4]), row[6]
            all_datum[hall_number] = {
                'choice': choice,
                'number': number,
                'hall_symbol': hall_symbol,
            }

    assert len(all_datum) == 530
    return all_datum


def get_msg_table():
    # Load MSG for ITA standard settings
    with open(Path("./magnetic_hall_symbols.yaml"), 'r') as f:
        all_datum = dict(YAML().load(f))
    return all_datum



#!/usr/bin/sage

import sys
import numpy as np
from sage.all import *

def get_VSpU( M ):
    N = copy(M)
    for i in range(N.nrows()):
        N[i,i%3] -= 1
    S,U,V = N.smith_form()

    Sp = Matrix( QQ, S.transpose() )
    for i in range(3):
        if not Sp[i,i]==0:
            Sp[i,i] = 1.0 / Sp[i,i]

    return V*Sp*U

def get_VSpU_sets(g1s, g2s):
    VSpU_sets = []
    generator_sets = []

    for g3 in ( False, True ): # for inversion
        for g1 in g1s:
            for g2 in g2s:
                if g2 is not False:
                    if np.equal(g1, g2).all():
                        continue
                    if np.equal(g1, np.dot(inv, g2)).all():
                        continue
                    genes = [g1, g2]
                else:
                    genes = [g1,]

                if g3:
                    genes_new = []
                    for g in genes:
                        if np.linalg.det(g) < 0:
                            genes_new.append(np.dot(inv, g))
                        else:
                            genes_new.append(g)
                    genes_new.append(inv)

                    if np.linalg.det(g1) < 0:
                        is_found = False
                        for g1_comp in g1s:
                            if np.equal(genes_new[0], g1_comp).all():
                                is_found = True
                                break
                        if is_found:
                            continue

                    if g2 is not False:
                        if np.linalg.det(g2) < 0:
                            is_found = False
                            for g2_comp in g2s:
                                if np.equal(genes_new[1], g2_comp).all():
                                    is_found = True
                                    break
                            if is_found:
                                continue

                else:
                    genes_new = genes

                M = Matrix(3, 3, genes_new[0])
                if len(genes_new) > 1:
                    for g in genes_new[1:]:
                        M = M.stack(Matrix(3, 3, g))
                VSpU_sets.append(get_VSpU(M))
                generator_sets.append( M )
    return VSpU_sets, generator_sets

def get_rotation_primitive( g1s, g2s, T ):
    if T is not None:
        g1s_new = []
        g2s_new = []
        g1s_old = []
        g2s_old = []
        for g in g1s:
            # print g, "-->"
            try:
                M = Matrix(ZZ,T*Matrix(QQ, g)*T.inverse())
                # print M
                g1s_new.append(np.array(M))
                g1s_old.append(g)
            except TypeError:
                print "Not integer matrix, pass this matrix"
        for g in g2s:
            if g:
                # print g, "-->"
                try:
                    M = Matrix(ZZ, T*Matrix(QQ, g)*T.inverse())
                    # print M
                    g2s_new.append(np.array(M))
                    g2s_old.append(g)
                except TypeError:
                    print "Not integer matrix, pass this matrix"
            else:
                g2s_new.append(False)
                g2s_old.append(False)
    return g1s_new, g2s_new, g1s_old, g2s_old

def write_generators(generator_sets):
    print "{"
    for count, M in enumerate( generator_sets ):
        print "  { /* %d */" % (count+1)
        for i in range( 3 ):
            print "    { ",
            for j in range( 3 ):
                for k in range( 3 ):
                    if M.nrows() // 3 > i:
                        print "%d," % M[i*3+j,k],
                    else:
                        print " 0,",
            print "},"
        print "  },"
    print "};"

def write_VSpU(VSpU_sets):
    print "{"
    for count, VSpU in enumerate( VSpU_sets ):
        print "  { /* %d */" % (count+1)
        for colvals in VSpU:
            print "    {",
            for i in range( 9 ):
                if i < len( colvals ):
                    num = colvals[i].numerator()
                    den = colvals[i].denominator()
                    if den == 1:
                        print "%d," % ( num ),
                    else:
                        print "%d.0/%d," % ( num, den ),
                else:
                    print " 0,",
            print "},"
        print "  },"
    print "};"


identity = [[ 1, 0, 0],
            [ 0, 1, 0],
            [ 0, 0, 1]]

inv = [[-1, 0, 0],
       [ 0,-1, 0],
       [ 0, 0,-1]]

# Tetragonal
rot4z = [[ 0,-1, 0],
         [ 1, 0, 0],
         [ 0, 0, 1]]
rot4zi = [[ 0, 1, 0],
          [-1, 0, 0],
          [ 0, 0,-1]]
rot2z = [[-1, 0, 0],
         [ 0,-1, 0],
         [ 0, 0, 1]]
rot2zi = [[ 1, 0, 0],
          [ 0, 1, 0],
          [ 0, 0,-1]]
rot2x = [[ 1, 0, 0],
         [ 0,-1, 0],
         [ 0, 0,-1]]
rot2xi = [[-1, 0, 0],
          [ 0, 1, 0],
          [ 0, 0, 1]]
rot3xyz = [[ 0, 0, 1],
           [ 1, 0, 0],
           [ 0, 1, 0]]
rot3xyzi = [[ 0, 0,-1],
            [-1, 0, 0],
            [ 0,-1, 0]]
hexa2_ab = [[ 0,-1, 0],
            [-1, 0, 0],
            [ 0, 0,-1]]
hexa2_abi = [[ 0, 1, 0],
             [ 1, 0, 0],
             [ 0, 0, 1]]
rot6z = [[ 1,-1, 0],
         [ 1, 0, 0],
         [ 0, 0, 1]]
rot6zi = [[-1, 1, 0],
          [-1, 0, 0],
          [ 0, 0,-1]]
trigo2ab = [[ 0, 1, 0],
            [ 1, 0, 0],
            [ 0, 0,-1]]
trigo2abi = [[ 0,-1, 0],
             [-1, 0, 0],
             [ 0, 0, 1]]
rot3z = [[ 0,-1, 0],
         [ 1,-1, 0],
         [ 0, 0, 1]]
rot3zi = [[ 0, 1, 0],
          [-1, 1, 0],
          [ 0, 0,-1]]
rot2y = [[-1, 0, 0],
         [ 0, 1, 0],
         [ 0, 0,-1]]
rot2yi = [[ 1, 0, 0],
          [ 0,-1, 0],
          [ 0, 0, 1]]



BCC = Matrix(QQ, [[ 0, 1, 1],
                  [ 1, 0, 1],
                  [ 1, 1, 0]])
FCC = Matrix(QQ, [[-1, 1, 1],
                  [ 1,-1, 1],
                  [ 1, 1,-1]])
CC = Matrix(QQ, [[ 1,-1, 0],
                 [ 1, 1, 0],
                 [ 0, 0, 1]])
BC = Matrix(QQ, [[ 1, 0, 1],
                 [ 0, 1, 0],
                 [-1, 0, 1]])
AC = Matrix(QQ, [[ 1, 0, 0],
                 [ 0, 1, 1],
                 [ 0,-1, 1]])
RC = Matrix(QQ, [[ 1, 0, 1],
                 [-1, 1, 1],
                 [ 0,-1, 1]])


# Parse options
from optparse import OptionParser
parser = OptionParser()
parser.set_defaults(is_tricli=False,
                    is_monocli=False,
                    is_ortho=False,
                    is_tetra=False,
                    is_rhombo=False,
                    is_trigo=False,
                    is_hexa=False,
                    is_cubic=False,
                    is_bcc=False,
                    is_fcc=False,
                    is_ac=False,
                    is_bc=False,
                    is_cc=False,
                    is_rc=False)
parser.add_option("--tricli", dest="is_tricli",
                  action="store_true")
parser.add_option("--monocli", dest="is_monocli",
                  action="store_true")
parser.add_option("--ortho", dest="is_ortho",
                  action="store_true")
parser.add_option("--tetra", dest="is_tetra",
                  action="store_true")
parser.add_option("--rhombo", dest="is_rhombo",
                  action="store_true")
parser.add_option("--trigo", dest="is_trigo",
                  action="store_true")
parser.add_option("--hexa", dest="is_hexa",
                  action="store_true")
parser.add_option("--cubic", dest="is_cubic",
                  action="store_true")
parser.add_option("--bcc", dest="is_bcc",
                  action="store_true")
parser.add_option("--fcc", dest="is_fcc",
                  action="store_true")
parser.add_option("--ac", dest="is_ac",
                  action="store_true")
parser.add_option("--bc", dest="is_bc",
                  action="store_true")
parser.add_option("--cc", dest="is_cc",
                  action="store_true")
parser.add_option("--rc", dest="is_rc",
                  action="store_true")
parser.add_option("-g", dest="is_generator",
                  action="store_true")
(options, args) = parser.parse_args()

g1s = None
g2s = None
g1s_old = None
g2s_old = None

if options.is_tricli:
    g1s = ( identity, )
    g2s = ( False, )

if options.is_monocli:
    g1s = ( rot2x, rot2y, rot2z, rot2xi, rot2yi, rot2zi )
    g2s = ( False, )
    if options.is_bcc:
        g1s, g2s, g1s_old, g2s_old = get_rotation_primitive( g1s, g2s, BCC )
    if options.is_ac:
        g1s, g2s, g1s_old, g2s_old = get_rotation_primitive( g1s, g2s, AC )
    if options.is_bc:
        g1s, g2s, g1s_old, g2s_old = get_rotation_primitive( g1s, g2s, BC )
    if options.is_cc:
        g1s, g2s, g1s_old, g2s_old = get_rotation_primitive( g1s, g2s, CC )

if options.is_ortho:
    g1s = ( rot2z, rot2zi )
    g2s = ( rot2x, rot2xi )
    if options.is_bcc:
        g1s, g2s, g1s_old, g2s_old = get_rotation_primitive( g1s, g2s, BCC )
    if options.is_fcc:
        g1s, g2s, g1s_old, g2s_old = get_rotation_primitive( g1s, g2s, FCC )
    if options.is_ac:
        g1s, g2s, g1s_old, g2s_old = get_rotation_primitive( g1s, g2s, AC )
    if options.is_bc:
        g1s, g2s, g1s_old, g2s_old = get_rotation_primitive( g1s, g2s, BC )
    if options.is_cc:
        g1s, g2s, g1s_old, g2s_old = get_rotation_primitive( g1s, g2s, CC )

if options.is_tetra:
    g1s = ( rot4z, rot4zi )
    g2s = ( False, rot2x, rot2xi )
    if options.is_bcc:
        g1s, g2s, g1s_old, g2s_old = get_rotation_primitive( g1s, g2s, BCC )

# if options.is_rhombo:
#     g1s = ( rot3xyz, rot3xyzi )
#     g2s = ( False, hexa2_ab, trigo2ab )

if options.is_trigo:
    g1s = ( rot3z, rot3zi )
    g2s = ( False, hexa2_ab, trigo2ab, hexa2_abi, trigo2abi )
    if options.is_rc: # hP
        g1s, g2s, g1s_old, g2s_old = get_rotation_primitive( g1s, g2s, RC )

if options.is_rhombo: # hR
    g1s = ( rot3z, rot3zi )
    g2s = ( False, hexa2_ab, trigo2ab, hexa2_abi, trigo2abi )
    g1s, g2s, g1s_old, g2s_old = get_rotation_primitive( g1s, g2s, RC )
    g1s_old = None
    g2s_old = None

if options.is_hexa:
    g1s = ( rot6z, rot6zi )
    g2s = ( False, hexa2_ab, hexa2_abi )

if options.is_cubic:
    g1s = ( rot4z, rot2z, rot4zi, rot2zi )
    g2s = ( rot3xyz, rot3xyzi )
    if options.is_bcc:
        g1s, g2s, g1s_old, g2s_old = get_rotation_primitive( g1s, g2s, BCC )
    if options.is_fcc:
        g1s, g2s, g1s_old, g2s_old = get_rotation_primitive( g1s, g2s, FCC )

if g1s is None:
    print "Option is needed. See \'make_VSpU.py -h\'"
else:
    if g1s_old is None:
        VSpU_sets, generator_sets = get_VSpU_sets( g1s, g2s )
    else:
        VSpU_sets = get_VSpU_sets( g1s, g2s )[0]
        generator_sets = get_VSpU_sets( g1s_old, g2s_old )[1]

    centering = ""
    if options.is_fcc:
        centering = "_F"
    if options.is_bcc:
        centering = "_I"
    if options.is_ac:
        centering = "_A"
    if options.is_bc:
        centering = "_B"
    if options.is_cc:
        centering = "_C"

    if options.is_generator:
        if options.is_tricli:
            print "static int tricli_generators[][3][9] ="
        if options.is_monocli:
            print "static int monocli_generators[][3][9] ="
        if options.is_ortho:
            print "static int ortho_generators[][3][9] ="
        if options.is_tetra:
            print "static int tetra_generators[][3][9] ="
        if options.is_trigo:
            if options.is_rc:
                print "static int rhombo_h_generators[][3][9] ="
            else:
                print "static int trigo_generators[][3][9] ="
        if options.is_rhombo:
            print "static int rhombo_p_generators[][3][9] ="
        if options.is_hexa:
            print "static int hexa_generators[][3][9] ="
        if options.is_cubic:
            print "static int cubic_generators[][3][9] ="
        write_generators(generator_sets)
        print
    else:
        if options.is_tricli:
            print "static double tricli_VSpU[][3][9] ="
        if options.is_monocli:
            print "static double monocli%s_VSpU[][3][9] =" % centering
        if options.is_ortho:
            print "static double ortho%s_VSpU[][3][9] =" % centering
        if options.is_tetra:
            print "static double tetra%s_VSpU[][3][9] =" % centering
        if options.is_trigo:
            if options.is_rc:
                print "static double rhombo_h_VSpU[][3][9] ="
            else:
                print "static double trigo_VSpU[][3][9] ="
        if options.is_rhombo:
                print "static double rhombo_p_VSpU[][3][9] ="
        if options.is_hexa:
                print "static double hexa_VSpU[][3][9] ="
        if options.is_cubic:
                print "static double cubic%s_VSpU[][3][9] ="
        write_VSpU(VSpU_sets)
        print



from itertools import combinations

import numpy as np
import pytest

from operation import MagneticOperation, remainder1_symmetry_operation
from magnetic_hall import MagneticHallSymbol
from transform import Transformation, get_standard_hall_number
from load import get_spg_table, get_msg_table


def test_magnetic_operations():
    tc = MagneticOperation.from_linear_translation_time_reversal(
        translation=np.array([0, 0, 1 / 2]), time_reversal=True
    )
    tc2 = tc * tc
    assert tc2 == MagneticOperation.from_linear_translation_time_reversal(
        translation=np.array([0, 0, 1])
    )

    # Translation by (0, 0, -1/2) is equivalent to that by (0, 0, 1/2)
    assert remainder1_symmetry_operation(tc.inverse()) == tc


@pytest.mark.parametrize(
    "hall_symbol,order_expected", [
        ("P 31 2 1c' (0 0 4)", 2 * 6),  # 151.32 (type-IV)
        ("P 6c 2c' -1'", 24),  # 194.265 (type-III)
        ("F 4d 2 3 1'", 2 * 96),  # 210.53 (type-II)
    ]
)
def test_magnetic_hall_symbol(hall_symbol, order_expected):
    mhs = MagneticHallSymbol(hall_symbol)
    order_actual = len(mhs.coset)
    assert order_actual == order_expected


def test_transformation():
    spg_table = get_spg_table()

    for hall_number in range(1, 530 + 1):
        print(hall_number)
        choice = spg_table[hall_number]['choice']
        number = spg_table[hall_number]['number']
        hall_symbol = spg_table[hall_number]['hall_symbol']
        coset_expected = MagneticHallSymbol(hall_symbol).coset

        hall_number_std = get_standard_hall_number(number)
        hall_symbol_std = spg_table[hall_number_std]['hall_symbol']
        coset_std = MagneticHallSymbol(hall_symbol_std).coset

        # Transform from ITA standard setting to alternative setting
        from_std = Transformation.to_standard(hall_number, choice, number).inverse()
        coset_actual = from_std.transform_coset(coset_std)

        assert set(coset_actual) == set(coset_expected)


def test_magnetic_hall_table():
    # Test MSG table with ITA standard settings
    msg_table = get_msg_table()

    count = 0
    for number in range(1, 230 + 1):
        print(f"No. {number}")
        hall_number_std = get_standard_hall_number(number)
        serial_and_symbols = list(msg_table[hall_number_std].items())

        hs_sg = MagneticHallSymbol(serial_and_symbols[0][1])  # type-I

        list_msg = []
        for bns_number, hall_symbol in serial_and_symbols:
            print(f"  {bns_number}: {hall_symbol}")
            mhs = MagneticHallSymbol(hall_symbol)
            assert len(mhs.coset) % len(hs_sg.coset) == 0

            index = len(mhs.coset) // len(hs_sg.coset)
            # For type-I and type-III, index == 1
            # For type-II and type-IV, index == 2
            assert (index == 1) or (index == 2)

            list_msg.append(set(mhs.coset))

            count += 1

        # check no duplicates
        for i, j in combinations(range(len(list_msg)), r=2):
            if list_msg[i] == list_msg[j]:
                print(serial_and_symbols[i], serial_and_symbols[j])
                assert False

    assert count == 1651

def test_all_msg_table():
    msg_table = get_msg_table()
    spg_table = get_spg_table()

    for hall_number in range(1, 530 + 1):
        number = spg_table[hall_number]['number']
        choice = spg_table[hall_number]['choice']
        hall_number_std = get_standard_hall_number(number)
        # Transformation from ITA standard settings to the other settings
        trns = Transformation.to_standard(hall_number, choice, number).inverse()

        for bns_number, mhall_symbol in msg_table[hall_number_std].items():
            mhs = MagneticHallSymbol(mhall_symbol)
            coset = trns.transform_coset(mhs.coset)

            # Check the order of coset
            assert len(mhs.coset) % len(coset) == 0
            if hall_number in [
                434,
                437,
                445,
                451,
                453,
                459,
                461,
            ]:
                # seven rhombohedral space groups
                assert len(mhs.coset) // len(coset) == 3
            else:
                assert len(mhs.coset) == len(coset)

            for op in coset:
                print(hall_number, bns_number, np.around(op.linear).astype(int))
                assert np.isclose(np.abs(np.linalg.det(op.linear)), 1)



import sys

import numpy as np

from operation import MagneticOperation
from magnetic_hall import MagneticHallSymbol
from transform import Transformation, get_standard_hall_number
from load import get_msg_numbers, get_spg_table, get_msg_table

sys.path.append('..')
from hall2operations import encode_symmetry


def get_all_msg_settings():
    msg_table = get_msg_table()
    spg_table = get_spg_table()

    msg_numbers = get_msg_numbers()
    mapping = {}
    for _, bns_number, _, uni_number in msg_numbers:
        mapping[bns_number] = uni_number

    all_datum = []
    offset = 1
    for hall_number in range(1, 530 + 1):
        number = spg_table[hall_number]['number']
        choice = spg_table[hall_number]['choice']
        hall_number_std = get_standard_hall_number(number)
        # Transformation from ITA standard settings to the other settings
        trns = Transformation.to_standard(hall_number, choice, number).inverse()

        for bns_number, mhall_symbol in msg_table[hall_number_std].items():
            mhs = MagneticHallSymbol(mhall_symbol)
            coset = trns.transform_coset(mhs.coset)
            uni_number = mapping[bns_number]
            all_datum.append((coset, uni_number, hall_number, offset))
            offset += len(coset)

    return all_datum


def encode_magnetic_operation(ops: MagneticOperation) -> int:
    linear = np.around(ops.linear).astype(int)
    assert np.all(np.abs(linear) <= 1)  # sanity check
    rt_encoded = encode_symmetry(linear, ops.translation)
    if ops.time_reversal:
        rt_encoded += 34012224  # 3**9 * 12**3

    assert rt_encoded < (2 ** 31)  # within int
    return rt_encoded


if __name__ == "__main__":
    all_datum = get_all_msg_settings()
    contents = []

    # Dump `magnetic_symmetry_operations`
    # Magnetic symmetry operations is encoded as well for time_reversal=False.
    # For time_reversal=True operation, encoded number is shifted.
    contents.append("static const int magnetic_symmetry_operations[] = {")
    contents.append("           0, /*dummy*/")
    count = 1
    for i, (coset, uni_number, _, _) in enumerate(all_datum):
        for ops in coset:
            encoded = encode_magnetic_operation(ops)
            linear = np.around(ops.linear).astype(int)
            linear_str = "%2d, %2d, %2d, %2d, %2d, %2d, %2d, %2d, %2d" % tuple(linear.reshape(-1).tolist())
            translation_str = "%2d, %2d, %2d" % tuple(np.around(12 * ops.translation))
            time_reversal_str = "-1" if ops.time_reversal else " 1"
            contents.append(f'    {encoded:8d}, /* {count:5d} ({uni_number:4d}) [{linear_str}] [{translation_str}] {time_reversal_str} */')
            count += 1
    contents.append('};')
    contents.append('')

    # Dump `magnetic_spacegroup_operation_index`
    # magnetic_spacegroup_operation_index[uni_number][18][2] = array of {order, offset} of the Hall number in `magnetic_symmetry_operations`
    mapping = {uni_number: [] for uni_number in range(1, 1651 + 1)}
    for coset, uni_number, hall_number, offset in all_datum:
        order = len(coset)
        mapping[uni_number].append((hall_number, order, offset))

    contents.append("static const int magnetic_spacegroup_operation_index[][18][2] = {")
    contents.append("    { {0, 0} }, /* dummy */")
    count = 1
    for uni_number in range(1, 1651 + 1):
        hall_order_offset = mapping[uni_number]
        assert len(hall_order_offset) <= 18

        orders_offsets = []
        for _, order, offset in hall_order_offset:
            orders_offsets.append((order, offset))

        orders_offsets_str = ", ".join([f"{{ {order}, {offset} }}" for order, offset in orders_offsets])
        contents.append(f"    {{ {orders_offsets_str} }}, /* {uni_number} */")
    contents.append("};")

    for line in contents:
        print(line)



from spglib import get_layergroup


def test_MoS2(root_data_dir, read_vasp):
    cell = read_vasp(root_data_dir / "layer" / "POSCAR-78")
    lg = get_layergroup(cell, aperiodic_dir=2, symprec=1e-5)
    assert lg.number == 78



#!/usr/bin/env python

import numpy as np

identity = [[ 1, 0, 0 ],
            [ 0, 1, 0 ],
            [ 0, 0, 1 ]]
permute_abc = [[ 0, 0, 1 ],
               [ 1, 0, 0 ],
               [ 0, 1, 0 ]]
invert_b = [[ 0, 0, 1 ],
            [ 0,-1, 0 ],
            [ 1, 0, 0 ]]
invert_c = [[ 0, 1, 0 ],
            [ 1, 0, 0 ],
            [ 0, 0,-1 ]]
monocli_cell_choice = [[-1, 0, 1 ],
                       [ 0, 1, 0 ],
                       [-1, 0, 0 ]]
monocli_centers = ['C', 'A', 'I', 'C']


def print_monocli():
    mat = identity
    transform_mats = []
    centerings = []
    center = 'C'
    for i in range(3):
        for j in range(3):  # changing choices
            for k in range(2):  # a <-> c, -b
                transform_mats.append(mat)
                centerings.append(center)

                mat = np.dot(mat, invert_b)
                if center == 'C':
                    center = 'A'
                elif center == 'A':
                    center = 'C'
                elif center == 'B':
                    center = 'B'
                elif center == 'I':
                    center = 'I'
            mat = np.dot(mat, permute_abc)
            if center == 'C':
                center = 'B'
            elif center == 'B':
                center = 'A'
            elif center == 'A':
                center = 'C'
            elif center == 'I':
                center = 'I'
        mat = np.dot(mat, monocli_cell_choice)
        center = monocli_centers[(i + 1) % 4]

    for mat in np.array(transform_mats):
        print(" {{%2d,%2d,%2d }," % tuple(mat[0]))
        print("  {%2d,%2d,%2d }," % tuple(mat[1]))
        print("  {%2d,%2d,%2d }}," % tuple(mat[2]))
    for center in centerings:
        if center == 'C':
            print("C_FACE,")
        if center == 'A':
            print("A_FACE,")
        if center == 'B':
            print("B_FACE,")
        if center == 'I':
            print("BODY,")


def print_ortho():
    mat = identity
    transform_mats = []
    centerings = []
    center = 'C'
    for i in range(2):
        for j in range(3):
            transform_mats.append(mat)
            centerings.append(center)
            mat = np.dot(mat, permute_abc)
            if center == 'C':
                center = 'B'
            elif center == 'B':
                center = 'A'
            elif center == 'A':
                center = 'C'
        mat = np.dot(mat, invert_c)
        if center == 'C':
            center = 'C'
        elif center == 'A':
            center = 'B'
        elif center == 'B':
            center = 'A'

    for mat in np.array(transform_mats):
        print(" {{%2d,%2d,%2d }," % tuple(mat[0]))
        print("  {%2d,%2d,%2d }," % tuple(mat[1]))
        print("  {%2d,%2d,%2d }}," % tuple(mat[2]))
    for center in centerings:
        if center == 'C':
            print("C_FACE,")
        if center == 'A':
            print("A_FACE,")
        if center == 'B':
            print("B_FACE,")
        if center == 'I':
            print("BODY,")


print_monocli()
# print_ortho()
